#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from typing import List

from marshmallow import EXCLUDE, ValidationError, fields, pre_load, validates_schema

from polyaxon.auxiliaries import (
    DefaultSchedulingSchema,
    PolyaxonCleanerSchema,
    PolyaxonInitContainerSchema,
    PolyaxonNotifierSchema,
    PolyaxonSidecarContainerSchema,
    V1DefaultScheduling,
)
from polyaxon.contexts import paths as ctx_paths
from polyaxon.env_vars.keys import (
    EV_KEYS_AGENT_ARTIFACTS_STORE,
    EV_KEYS_AGENT_CLEANER,
    EV_KEYS_AGENT_COMPRESSED_LOGS,
    EV_KEYS_AGENT_CONNECTIONS,
    EV_KEYS_AGENT_DEFAULT_IMAGE_PULL_SECRETS,
    EV_KEYS_AGENT_DEFAULT_SCHEDULING,
    EV_KEYS_AGENT_INIT,
    EV_KEYS_AGENT_IS_REPLICA,
    EV_KEYS_AGENT_NOTIFIER,
    EV_KEYS_AGENT_RUNS_SA,
    EV_KEYS_AGENT_SECRET_NAME,
    EV_KEYS_AGENT_SIDECAR,
    EV_KEYS_AGENT_SPAWNER_REFRESH_INTERVAL,
    EV_KEYS_AGENT_USE_PROXY_ENV_VARS_IN_OPS,
    EV_KEYS_K8S_APP_SECRET_NAME,
    EV_KEYS_K8S_NAMESPACE,
    EV_KEYS_SANDBOX_DEBUG,
    EV_KEYS_SANDBOX_HOST,
    EV_KEYS_SANDBOX_PER_CORE,
    EV_KEYS_SANDBOX_PORT,
    EV_KEYS_SANDBOX_SSL_ENABLED,
    EV_KEYS_SANDBOX_WORKERS,
)
from polyaxon.exceptions import PolyaxonSchemaError
from polyaxon.lifecycle import V1ProjectFeature
from polyaxon.parser import parser
from polyaxon.schemas.base import BaseConfig, BaseSchema
from polyaxon.schemas.types import ConnectionTypeSchema, V1K8sResourceType
from polyaxon.utils.http_utils import clean_host
from polyaxon.utils.signal_decorators import check_partial


def validate_agent_config(
    artifacts_store, connections, required_artifacts_store: bool = True
):
    if required_artifacts_store and not artifacts_store:
        raise ValidationError(
            "A connection definition is required to set a default artifacts store."
        )

    connections = connections or []

    connection_names = set()

    if artifacts_store:
        connection_names.add(artifacts_store.name)

    for c in connections:
        if c.name in connection_names:
            raise ValidationError(
                "A connection with name `{}` must be unique.".format(c.name)
            )
        connection_names.add(c.name)


class BaseAgentSchema(BaseSchema):
    REQUIRED_ARTIFACTS_STORE = True

    artifacts_store = fields.Nested(
        ConnectionTypeSchema,
        allow_none=True,
        data_key=EV_KEYS_AGENT_ARTIFACTS_STORE,
    )
    connections = fields.List(
        fields.Nested(ConnectionTypeSchema),
        allow_none=True,
        data_key=EV_KEYS_AGENT_CONNECTIONS,
    )

    @validates_schema
    @check_partial
    def validate_connection(self, data, **kwargs):
        validate_agent_config(
            data.get("artifacts_store"),
            data.get("connections"),
            self.REQUIRED_ARTIFACTS_STORE,
        )

    @pre_load
    def pre_validate(self, data, **kwargs):
        connections = data.get(EV_KEYS_AGENT_CONNECTIONS)
        try:
            connections = parser.get_dict(
                key=EV_KEYS_AGENT_CONNECTIONS,
                value=connections,
                is_list=True,
                is_optional=True,
            )
        except PolyaxonSchemaError as e:
            raise ValidationError("Received an invalid connections") from e
        if connections:
            data[EV_KEYS_AGENT_CONNECTIONS] = connections

        artifacts_store = data.get(EV_KEYS_AGENT_ARTIFACTS_STORE)
        try:
            artifacts_store = parser.get_dict(
                key=EV_KEYS_AGENT_ARTIFACTS_STORE,
                value=artifacts_store,
                is_optional=True,
            )
        except PolyaxonSchemaError as e:
            raise ValidationError(
                "Received an invalid artifacts store `{}`".format(artifacts_store)
            ) from e
        if artifacts_store:
            data[EV_KEYS_AGENT_ARTIFACTS_STORE] = artifacts_store

        return data


class BaseAgentConfig(BaseConfig):
    UNKNOWN_BEHAVIOUR = EXCLUDE
    REDUCED_ATTRIBUTES = [
        EV_KEYS_AGENT_ARTIFACTS_STORE,
        EV_KEYS_AGENT_CONNECTIONS,
    ]

    def __init__(
        self, artifacts_store=None, connections=None, namespace: str = None, **kwargs
    ):
        self.namespace = namespace
        self.artifacts_store = artifacts_store
        self.connections = connections or []
        self._all_connections = []
        self.set_all_connections()
        self._secrets = None
        self._config_maps = None
        self._connections_by_names = {}

    def set_all_connections(self):
        self._all_connections = self.connections[:]
        if self.artifacts_store:
            self._all_connections.append(self.artifacts_store)
            validate_agent_config(self.artifacts_store, self.connections)

    @property
    def all_connections(self):
        return self._all_connections

    @property
    def secrets(self) -> List[V1K8sResourceType]:
        if self._secrets or not self._all_connections:
            return self._secrets
        secret_names = set()
        secrets = []
        for c in self._all_connections:
            if c.secret and c.secret.name not in secret_names:
                secret_names.add(c.secret.name)
                secrets.append(c.get_secret())
        self._secrets = secrets
        return self._secrets

    @property
    def config_maps(self) -> List[V1K8sResourceType]:
        if self._config_maps or not self._all_connections:
            return self._config_maps
        config_map_names = set()
        config_maps = []
        for c in self._all_connections:
            if c.config_map and c.config_map.name not in config_map_names:
                config_map_names.add(c.config_map.name)
                config_maps.append(c.get_config_map())
        self._config_maps = config_maps
        return self._config_maps

    @property
    def connections_by_names(self):
        if self._connections_by_names or not self._all_connections:
            return self._connections_by_names

        self._connections_by_names = {c.name: c for c in self._all_connections}
        return self._connections_by_names

    @property
    def local_root(self):
        artifacts_root = ctx_paths.CONTEXT_ARTIFACTS_ROOT
        if not self.artifacts_store:
            return artifacts_root

        if self.artifacts_store.is_mount:
            return self.artifacts_store.store_path

        return artifacts_root

    def get_local_path(self, subpath: str, entity: str = None):
        full_path = self.local_root
        if entity == V1ProjectFeature.RUNTIME:
            from polyaxon.services.values import PolyaxonServices

            if PolyaxonServices.is_sandbox():
                full_path = os.path.join(full_path, "runs")
        else:
            full_path = os.path.join(full_path, f"{entity}s")

        return f"{full_path}/{subpath}"

    @property
    def store_root(self):
        artifacts_root = ctx_paths.CONTEXT_ARTIFACTS_ROOT
        if not self.artifacts_store:
            return artifacts_root

        return self.artifacts_store.store_path

    def get_store_path(self, subpath: str, entity: str = None):
        full_path = self.store_root
        if entity == V1ProjectFeature.RUNTIME:
            from polyaxon.services.values import PolyaxonServices

            if PolyaxonServices.is_sandbox():
                full_path = os.path.join(full_path, "runs")
        else:
            full_path = os.path.join(full_path, f"{entity}s")

        if subpath:
            full_path = os.path.join(full_path, subpath)
        return full_path


class SandboxSchema(BaseAgentSchema):
    REQUIRED_ARTIFACTS_STORE = False

    port = fields.Int(allow_none=True, data_key=EV_KEYS_SANDBOX_PORT)
    host = fields.Str(allow_none=True, data_key=EV_KEYS_SANDBOX_HOST)
    ssl_enabled = fields.Bool(allow_none=True, data_key=EV_KEYS_SANDBOX_SSL_ENABLED)
    debug = fields.Bool(allow_none=True, data_key=EV_KEYS_SANDBOX_DEBUG)
    workers = fields.Int(allow_none=True, data_key=EV_KEYS_SANDBOX_WORKERS)
    per_core = fields.Bool(allow_none=True, data_key=EV_KEYS_SANDBOX_PER_CORE)

    @staticmethod
    def schema_config():
        return SandboxConfig


class SandboxConfig(BaseAgentConfig):
    SCHEMA = SandboxSchema
    IDENTIFIER = "sandbox"
    REDUCED_ATTRIBUTES = BaseAgentConfig.REDUCED_ATTRIBUTES + [
        EV_KEYS_SANDBOX_PORT,
        EV_KEYS_SANDBOX_HOST,
        EV_KEYS_SANDBOX_SSL_ENABLED,
        EV_KEYS_SANDBOX_DEBUG,
        EV_KEYS_SANDBOX_WORKERS,
        EV_KEYS_SANDBOX_PER_CORE,
    ]

    def __init__(
        self,
        artifacts_store=None,
        connections=None,
        port: int = None,
        host: str = None,
        ssl_enabled: bool = None,
        debug: bool = None,
        workers: int = None,
        per_core: bool = None,
        **kwargs,
    ):
        super().__init__(
            artifacts_store=artifacts_store,
            connections=connections,
            namespace="sandbox",
            **kwargs,
        )
        self.host = clean_host(host) if host else host
        self.port = port
        self.ssl_enabled = ssl_enabled
        self.debug = debug
        self.workers = workers
        self.per_core = per_core


class AgentSchema(BaseAgentSchema):
    REQUIRED_ARTIFACTS_STORE = True

    namespace = fields.Str(allow_none=True, data_key=EV_KEYS_K8S_NAMESPACE)
    is_replica = fields.Bool(allow_none=True, data_key=EV_KEYS_AGENT_IS_REPLICA)
    compressed_logs = fields.Bool(
        allow_none=True, data_key=EV_KEYS_AGENT_COMPRESSED_LOGS
    )
    sidecar = fields.Nested(
        PolyaxonSidecarContainerSchema,
        allow_none=True,
        data_key=EV_KEYS_AGENT_SIDECAR,
    )
    init = fields.Nested(
        PolyaxonInitContainerSchema, allow_none=True, data_key=EV_KEYS_AGENT_INIT
    )
    notifier = fields.Nested(
        PolyaxonNotifierSchema, allow_none=True, data_key=EV_KEYS_AGENT_NOTIFIER
    )
    cleaner = fields.Nested(
        PolyaxonCleanerSchema, allow_none=True, data_key=EV_KEYS_AGENT_CLEANER
    )
    use_proxy_env_vars_use_in_ops = fields.Bool(
        allow_none=True, data_key=EV_KEYS_AGENT_USE_PROXY_ENV_VARS_IN_OPS
    )
    default_scheduling = fields.Nested(
        DefaultSchedulingSchema,
        allow_none=True,
        data_key=EV_KEYS_AGENT_DEFAULT_SCHEDULING,
    )
    default_image_pull_secrets = fields.List(
        fields.Str(),
        allow_none=True,
        data_key=EV_KEYS_AGENT_DEFAULT_IMAGE_PULL_SECRETS,
    )
    app_secret_name = fields.Str(
        allow_none=True,
        data_key=EV_KEYS_K8S_APP_SECRET_NAME,
    )
    agent_secret_name = fields.Str(
        allow_none=True,
        data_key=EV_KEYS_AGENT_SECRET_NAME,
    )
    runs_sa = fields.Str(
        allow_none=True,
        data_key=EV_KEYS_AGENT_RUNS_SA,
    )
    # This refresh logic will mitigate several issues with AKS's numerous networking problems
    spawner_refresh_interval = fields.Integer(
        allow_none=True,
        data_key=EV_KEYS_AGENT_SPAWNER_REFRESH_INTERVAL,
    )

    @staticmethod
    def schema_config():
        return AgentConfig

    @pre_load
    def pre_validate(self, data, **kwargs):
        data = super().pre_validate(data, **kwargs)

        sidecar = data.get(EV_KEYS_AGENT_SIDECAR)
        try:
            sidecar = parser.get_dict(
                key=EV_KEYS_AGENT_SIDECAR, value=sidecar, is_optional=True
            )
        except PolyaxonSchemaError as e:
            raise ValidationError(
                "Received an invalid sidecar `{}`".format(sidecar)
            ) from e
        if sidecar:
            data[EV_KEYS_AGENT_SIDECAR] = sidecar

        init = data.get(EV_KEYS_AGENT_INIT)
        try:
            init = parser.get_dict(key=EV_KEYS_AGENT_INIT, value=init, is_optional=True)
        except PolyaxonSchemaError as e:
            raise ValidationError("Received an invalid init `{}`".format(init)) from e
        if init:
            data[EV_KEYS_AGENT_INIT] = init

        cleaner = data.get(EV_KEYS_AGENT_CLEANER)
        try:
            cleaner = parser.get_dict(
                key=EV_KEYS_AGENT_CLEANER, value=cleaner, is_optional=True
            )
        except PolyaxonSchemaError as e:
            raise ValidationError(
                "Received an invalid cleaner `{}`".format(cleaner)
            ) from e
        if cleaner:
            data[EV_KEYS_AGENT_CLEANER] = cleaner

        notifier = data.get(EV_KEYS_AGENT_NOTIFIER)
        try:
            notifier = parser.get_dict(
                key=EV_KEYS_AGENT_NOTIFIER, value=notifier, is_optional=True
            )
        except PolyaxonSchemaError as e:
            raise ValidationError(
                "Received an invalid notifier `{}`".format(notifier)
            ) from e
        if notifier:
            data[EV_KEYS_AGENT_NOTIFIER] = notifier

        default_scheduling = data.get(EV_KEYS_AGENT_DEFAULT_SCHEDULING)
        try:
            default_scheduling = parser.get_dict(
                key=EV_KEYS_AGENT_DEFAULT_SCHEDULING,
                value=default_scheduling,
                is_optional=True,
            )
        except PolyaxonSchemaError as e:
            raise ValidationError(
                "Received an invalid default_scheduling `{}`".format(default_scheduling)
            ) from e
        if default_scheduling:
            data[EV_KEYS_AGENT_DEFAULT_SCHEDULING] = default_scheduling

        default_image_pull_secrets = data.get(EV_KEYS_AGENT_DEFAULT_IMAGE_PULL_SECRETS)
        try:
            default_image_pull_secrets = parser.get_string(
                key=EV_KEYS_AGENT_DEFAULT_IMAGE_PULL_SECRETS,
                value=default_image_pull_secrets,
                is_optional=True,
                is_list=True,
            )
        except PolyaxonSchemaError as e:
            raise ValidationError(
                "Received an invalid default_image_pull_secrets `{}`".format(
                    default_image_pull_secrets
                )
            ) from e
        if default_image_pull_secrets:
            data[EV_KEYS_AGENT_DEFAULT_IMAGE_PULL_SECRETS] = default_image_pull_secrets

        return data


class AgentConfig(BaseAgentConfig):
    SCHEMA = AgentSchema
    IDENTIFIER = "agent"
    REDUCED_ATTRIBUTES = BaseAgentConfig.REDUCED_ATTRIBUTES + [
        EV_KEYS_AGENT_SIDECAR,
        EV_KEYS_AGENT_INIT,
        EV_KEYS_AGENT_NOTIFIER,
        EV_KEYS_AGENT_CLEANER,
        EV_KEYS_AGENT_IS_REPLICA,
        EV_KEYS_AGENT_COMPRESSED_LOGS,
        EV_KEYS_K8S_APP_SECRET_NAME,
        EV_KEYS_AGENT_SECRET_NAME,
        EV_KEYS_AGENT_RUNS_SA,
        EV_KEYS_AGENT_SPAWNER_REFRESH_INTERVAL,
        EV_KEYS_AGENT_USE_PROXY_ENV_VARS_IN_OPS,
        EV_KEYS_AGENT_DEFAULT_SCHEDULING,
        EV_KEYS_AGENT_DEFAULT_IMAGE_PULL_SECRETS,
    ]

    def __init__(
        self,
        namespace=None,
        is_replica=None,
        compressed_logs=None,
        sidecar=None,
        init=None,
        notifier=None,
        cleaner=None,
        artifacts_store=None,
        connections=None,
        app_secret_name=None,
        agent_secret_name=None,
        runs_sa=None,
        spawner_refresh_interval=None,
        default_scheduling=None,
        use_proxy_env_vars_use_in_ops=None,
        default_image_pull_secrets=None,
        **kwargs,
    ):
        super().__init__(
            artifacts_store=artifacts_store,
            connections=connections,
            namespace=namespace,
            **kwargs,
        )
        self.is_replica = is_replica
        self.compressed_logs = compressed_logs
        self.sidecar = sidecar
        self.init = init
        self.notifier = notifier
        self.cleaner = cleaner
        self.app_secret_name = app_secret_name
        self.agent_secret_name = agent_secret_name
        self.runs_sa = runs_sa
        self.spawner_refresh_interval = spawner_refresh_interval
        self.default_image_pull_secrets = default_image_pull_secrets
        self.use_proxy_env_vars_use_in_ops = use_proxy_env_vars_use_in_ops
        self.default_scheduling = default_scheduling
        if not self.default_scheduling and self.default_image_pull_secrets:
            self.default_scheduling = V1DefaultScheduling()
        if self.default_scheduling and not self.default_scheduling.image_pull_secrets:
            self.default_scheduling.image_pull_secrets = self.default_image_pull_secrets

    def get_spawner_refresh_interval(self):
        return self.spawner_refresh_interval or 60 * 5



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Union

from polyaxon.k8s import k8s_schemas


def has_tpu_annotation(annotations: Dict) -> bool:
    if not annotations:
        return False
    for key in annotations.keys():
        if "tpu" in key:
            return True

    return False


def requests_tpu(resources: Union[k8s_schemas.V1ResourceRequirements, Dict]) -> bool:
    if not resources:
        return False

    if not isinstance(resources, k8s_schemas.V1ResourceRequirements):
        resources = k8s_schemas.V1ResourceRequirements(**resources)

    if resources.requests:
        for key in resources.requests.keys():
            if "tpu" in key:
                return True

    if resources.limits:
        for key in resources.limits.keys():
            if "tpu" in key:
                return True

    return False


def requests_gpu(resources: k8s_schemas.V1ResourceRequirements) -> bool:
    if not resources:
        return False

    if resources.requests:
        for key in resources.requests.keys():
            if "gpu" in key:
                return True

    if resources.limits:
        for key in resources.limits.keys():
            if "gpu" in key:
                return True

    return False



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.deploy.operators.cmd_operator import CmdOperator
from polyaxon.exceptions import PolyaxonOperatorException


class DockerOperator(CmdOperator):
    CMD = "docker"

    @classmethod
    def params(cls, args):
        params = [cls.CMD] + args
        return params

    @classmethod
    def check(cls):
        command_exist = cls.execute(args=["version"])
        if not command_exist:
            return False
        return True

    @classmethod
    def execute(cls, args, stream=False):
        params = cls.params(args)
        return cls._execute(params=params, env=None, stream=stream)

    @classmethod
    def set_volume(cls, volume):
        args = ["volume", "create", "--name={}".format(volume)]
        if not volume:
            raise PolyaxonOperatorException(
                "docker", args, None, None, "Volume name was not provided"
            )
        return cls.execute(args)



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import ujson

from polyaxon.logger import logger


@click.group()
def tuner():
    logger.info("Creating new suggestions...")
    pass


@tuner.command()
@click.option(
    "--matrix",
    help="A string representing the matrix configuration for bayesian optimization.",
)
@click.option(
    "--configs",
    help="A string representing the list of configs.",
)
@click.option(
    "--metrics",
    help="A string representing the list of metrics.",
)
@click.option("--iteration", type=int, help="The current iteration.")
def bayes(matrix, configs, metrics, iteration):
    """Create suggestions based on bayesian optimization."""
    from hypertune.iteration_lineage import handle_iteration, handle_iteration_failure
    from hypertune.search_managers.bayesian_optimization.manager import (
        BayesSearchManager,
    )
    from polyaxon.client import RunClient
    from polyaxon.polyflow import V1Bayes

    matrix = V1Bayes.read(matrix)
    if configs:
        configs = ujson.loads(configs)
    if metrics:
        metrics = ujson.loads(metrics)

    client = RunClient()

    retry = 1
    exp = None
    suggestions = None
    while retry < 3:
        try:
            suggestions = BayesSearchManager(
                config=matrix,
            ).get_suggestions(configs=configs, metrics=metrics)
            exp = None
            break
        except Exception as e:
            retry += 1
            logger.warning(e)
            exp = e

    if exp:
        handle_iteration_failure(client=client, exp=exp)
        return

    handle_iteration(
        client=client,
        suggestions=suggestions,
    )


@tuner.command()
@click.option(
    "--matrix", help="A string representing the matrix configuration for hyperband."
)
@click.option(
    "--configs",
    help="A string representing the list of configs.",
)
@click.option(
    "--metrics",
    help="A string representing the list of metrics.",
)
@click.option("--iteration", type=int, help="The current hyperband iteration.")
@click.option(
    "--bracket-iteration",
    "--bracket_iteration",
    type=int,
    help="The current hyperband bracket iteration.",
)
def hyperband(matrix, configs, metrics, iteration, bracket_iteration):
    """Create suggestions based on hyperband."""
    from hypertune.iteration_lineage import handle_iteration, handle_iteration_failure
    from hypertune.search_managers.hyperband.manager import HyperbandManager
    from polyaxon.client import RunClient
    from polyaxon.polyflow import V1Hyperband

    matrix = V1Hyperband.read(matrix)
    matrix.set_tuning_params()
    if configs:
        configs = ujson.loads(configs)
    if metrics:
        metrics = ujson.loads(metrics)

    client = RunClient()

    retry = 1
    exp = None
    suggestions = None
    while retry < 3:
        try:
            suggestions = HyperbandManager(config=matrix).get_suggestions(
                configs=configs,
                metrics=metrics,
                bracket_iteration=bracket_iteration,
                iteration=iteration,
            )
            exp = None
            break
        except Exception as e:
            retry += 1
            logger.warning(e)
            exp = e

    if exp:
        handle_iteration_failure(client=client, exp=exp)
        return

    handle_iteration(
        client=client,
        suggestions=suggestions,
    )


@tuner.command()
@click.option(
    "--matrix", help="A string representing the matrix configuration for hyperopt."
)
@click.option(
    "--configs",
    help="A string representing the list of configs.",
)
@click.option(
    "--metrics",
    help="A string representing the list of metrics.",
)
@click.option("--iteration", type=int, help="The current iteration.")
def hyperopt(matrix, configs, metrics, iteration):
    """Create suggestions based on hyperopt."""
    from hypertune.iteration_lineage import handle_iteration, handle_iteration_failure
    from hypertune.search_managers.hyperopt.manager import HyperoptManager
    from polyaxon.client import RunClient
    from polyaxon.polyflow import V1Hyperopt

    matrix = V1Hyperopt.read(matrix)
    if configs:
        configs = ujson.loads(configs)
    if metrics:
        metrics = ujson.loads(metrics)

    client = RunClient()

    retry = 1
    exp = None
    suggestions = None
    while retry < 3:
        try:
            suggestions = HyperoptManager(config=matrix).get_suggestions(
                configs=configs, metrics=metrics
            )
            exp = None
            break
        except Exception as e:
            retry += 1
            logger.warning(e)
            exp = e

    if exp:
        handle_iteration_failure(client=client, exp=exp)
        return

    handle_iteration(
        client=client,
        suggestions=suggestions,
    )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def clean_version_for_compatibility(version: str):
    return "-".join(version.lstrip("v").replace(".", "-").split("-")[:3])


def clean_version_for_check(version: str):
    if not version:
        return version
    return ".".join(version.lstrip("v").replace("-", ".").split(".")[:3])


def compare_versions(current: str, reference: str, comparator: str) -> bool:

    from distutils.version import LooseVersion

    current = LooseVersion(current)
    reference = LooseVersion(reference)

    if comparator == "=":
        return current == reference

    if comparator == "!=":
        return current != reference

    if comparator == "<":
        return current < reference

    if comparator == "<=":
        return current <= reference

    if comparator == ">":
        return current > reference

    if comparator == ">=":
        return current >= reference

    raise ValueError("Comparator `{}` is not supported.".format(comparator))



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from marshmallow import fields

from polyaxon.schemas.base import BaseCamelSchema, BaseConfig


class IntervalsSchema(BaseCamelSchema):
    runs_scheduler = fields.Int(allow_none=True)
    operations_default_retry_delay = fields.Int(allow_none=True)
    operations_max_retry_delay = fields.Int(allow_none=True)
    compatibility_check = fields.Int(allow_none=True)

    @staticmethod
    def schema_config():
        return IntervalsConfig


class IntervalsConfig(BaseConfig):
    SCHEMA = IntervalsSchema
    REDUCED_ATTRIBUTES = [
        "runsScheduler",
        "operationsDefaultRetryDelay",
        "operationsMaxRetryDelay",
        "compatibilityCheck",
    ]

    def __init__(
        self,  # noqa
        runs_scheduler=None,
        operations_default_retry_delay=None,
        operations_max_retry_delay=None,
        compatibility_check=None,
    ):
        self.runs_scheduler = runs_scheduler
        self.operations_default_retry_delay = operations_default_retry_delay
        self.operations_max_retry_delay = operations_max_retry_delay
        self.compatibility_check = compatibility_check



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from typing import Optional

from polyaxon.utils.fqn_utils import to_fqn_name
from traceml.artifacts import V1ArtifactKind, V1RunArtifact


def collect_lineage_artifacts_path(artifact_path: str) -> Optional[V1RunArtifact]:
    name = os.path.basename(artifact_path)
    return V1RunArtifact(
        name=to_fqn_name(name),
        kind=V1ArtifactKind.DIR,
        path=artifact_path,
        summary={"path": artifact_path},
        is_input=True,
    )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import os

from kubernetes.client.rest import ApiException

from polyaxon.client import RunClient
from polyaxon.contexts import paths as ctx_paths
from polyaxon.env_vars.getters import get_run_info
from polyaxon.env_vars.keys import EV_KEYS_K8S_POD_ID
from polyaxon.exceptions import PolyaxonClientException, PolyaxonContainerException
from polyaxon.fs.fs import (
    close_fs,
    get_artifacts_connection_type,
    get_async_fs_from_type,
)
from polyaxon.fs.watcher import FSWatcher
from polyaxon.k8s.async_manager import AsyncK8SManager
from polyaxon.logger import logger
from polyaxon.settings import CLIENT_CONFIG
from polyaxon.sidecar.container.intervals import get_sync_interval
from polyaxon.sidecar.container.monitors import sync_artifacts, sync_logs
from polyaxon.sidecar.ignore import IGNORE_FOLDERS
from polyaxon.utils.tz_utils import now


async def start_sidecar(
    container_id: str,
    sleep_interval: int,
    sync_interval: int,
    monitor_outputs: bool,
    monitor_logs: bool,
):
    sync_interval = get_sync_interval(
        interval=sync_interval, sleep_interval=sleep_interval
    )
    try:
        pod_id = os.environ[EV_KEYS_K8S_POD_ID]
    except KeyError as e:
        raise PolyaxonContainerException(
            "Please make sure that this job has been "
            "started by Polyaxon with all required context."
        ) from e

    try:
        owner, project, run_uuid = get_run_info()
    except PolyaxonClientException as e:
        raise PolyaxonContainerException(e)

    client = RunClient(owner=owner, project=project, run_uuid=run_uuid)
    k8s_manager = AsyncK8SManager(namespace=CLIENT_CONFIG.namespace, in_cluster=True)
    await k8s_manager.setup()
    pod = await k8s_manager.get_pod(pod_id, reraise=True)
    connection_type = get_artifacts_connection_type()
    fs = await get_async_fs_from_type(connection_type=connection_type)
    fw = FSWatcher.read(ctx_paths.CONTEXT_MOUNT_FILE_WATCHER)

    retry = 1
    is_running = True
    counter = 0
    state = {
        "last_artifacts_check": None,
        "last_logs_check": None,
    }

    async def monitor():
        if monitor_logs:
            await sync_logs(
                run_uuid=run_uuid,
                k8s_manager=k8s_manager,
                pod=pod,
                last_time=None,
                stream=True,
                is_running=is_running,
            )
        if monitor_outputs:
            last_check = state["last_artifacts_check"]
            await sync_artifacts(
                fs=fs,
                fw=fw,
                store_path=connection_type.store_path,
                run_uuid=run_uuid,
                exclude=IGNORE_FOLDERS,
            )
            try:
                client.sync_events_summaries(
                    last_check=last_check,
                    events_path=ctx_paths.CONTEXT_MOUNT_RUN_EVENTS_FORMAT.format(
                        run_uuid
                    ),
                )
                client.sync_system_events_summaries(
                    last_check=last_check,
                    events_path=ctx_paths.CONTEXT_MOUNT_RUN_SYSTEM_RESOURCES_EVENTS_FORMAT.format(
                        run_uuid
                    ),
                )
                update_last_check = True
            except Exception as e:
                logger.info(
                    "An error occurred while syncing events summaries, "
                    "Exception %s" % repr(e)
                )
                update_last_check = False
            if update_last_check:
                state["last_artifacts_check"] = now()

    while is_running and retry <= 3:
        await asyncio.sleep(sleep_interval)
        try:
            is_running = await k8s_manager.is_pod_running(pod_id, container_id)
        except ApiException as e:
            retry += 1
            logger.info("Exception %s" % repr(e))
            logger.info("Sleeping ...")
            await asyncio.sleep(retry)
            continue

        logger.debug("Syncing ...")
        if is_running:
            retry = 1

        counter += 1
        if counter == sync_interval:
            counter = 0
            try:
                await monitor()
            except Exception as e:
                logger.warning("Polyaxon sidecar error: %s" % repr(e))

    await monitor()
    logger.info("Cleaning non main containers")
    if k8s_manager:
        await k8s_manager.close()

    await close_fs(fs)



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from polyaxon.connections.kinds import V1ConnectionKind
from polyaxon.notifiers.keys import INTEGRATIONS_WEBHOOKS_HIPCHAT
from polyaxon.notifiers.spec import NotificationSpec
from polyaxon.notifiers.webhook import WebHookNotifier


class HipChatWebHookNotifier(WebHookNotifier):
    notification_key = V1ConnectionKind.HIPCHAT
    name = "HipChat WebHook"
    description = "HipChat webhooks to send payload to a hipchat room."
    raise_empty_context = True
    config_key = INTEGRATIONS_WEBHOOKS_HIPCHAT

    @classmethod
    def serialize_notification_to_context(cls, notification: NotificationSpec) -> Dict:
        payload = {
            "message": notification.get_details(),
            "message_format": "text",
            "color": notification.get_color(),
            "from": "Polyaxon",
        }

        return payload

    @classmethod
    def _prepare(cls, context: Dict) -> Dict:
        context = super()._prepare(context)

        return {
            "message": context.get("message"),
            "message_format": context.get("message_format", "html"),
            "color": context.get("color"),
            "from": context.get("from", "Polyaxon"),
            "attach_to": context.get("attach_to"),
            "notify": context.get("notify", False),
            "card": context.get("card"),
        }



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.contexts import paths as ctx_paths
from polyaxon.fs.watcher import FSWatcher
from polyaxon.logger import logger
from polyaxon.schemas.types import V1ConnectionType
from polyaxon.utils.formatting import Printer


def sync_file_watcher(path: str):
    try:
        fw = FSWatcher()
        fw.sync(path)
        fw.write(ctx_paths.CONTEXT_MOUNT_FILE_WATCHER)
    except Exception as e:  # File watcher should not prevent job from starting
        logger.warning(
            "File watcher failed syncing path: {}.\nError: {}".format(path, e)
        )


def download_artifact(
    connection_name: str,
    connection_kind: str,
    path_from: str,
    path_to: str,
    is_file: bool,
    raise_errors: bool,
    sync_fw: bool,
    check_path: bool,
):
    from polyaxon.fs.fs import get_sync_fs_from_type
    from polyaxon.fs.manager import download_file_or_dir

    connection_type = V1ConnectionType(name=connection_name, kind=connection_kind)
    fs = get_sync_fs_from_type(connection_type=connection_type)

    try:
        download_file_or_dir(
            fs=fs,
            path_from=path_from,
            path_to=path_to,
            is_file=is_file,
            check_path=check_path,
        )
        if sync_fw:
            sync_file_watcher(path_to)
        Printer.print_success(
            "{} path is initialized, path: `{}`".format(connection_kind, path_to)
        )
    except Exception as e:
        if raise_errors:
            raise e
        else:
            logger.debug(
                "Initialization failed, the error was ignored. Error details %s", e
            )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ujson

from polyaxon.utils.formatting import Printer, dict_tabulate, dict_to_tabulate


def get_entity_details(entity: any, entity_name: str):
    if entity.description:
        Printer.print_heading("{} description:".format(entity_name))
        Printer.print("{}\n".format(entity.description))

    if entity.settings:
        Printer.print_heading("{} settings:".format(entity_name))
        Printer.print(
            "{}\n".format(
                entity.settings.to_dict()
                if hasattr(entity.settings, "to_dict")
                else entity.settings
            )
        )

    if entity.readme:
        Printer.print_heading("{} readme:".format(entity_name))
        Printer.print_md(entity.readme)

    response = dict_to_tabulate(
        entity.to_dict(),
        humanize_values=True,
        exclude_attrs=["description", "settings", "readme"],
    )

    Printer.print_heading("{} info:".format(entity_name))
    dict_tabulate(response)


def handle_output(response: any, output: str):
    if output == "json":
        Printer.pprint(response)
        return
    if "path=" in output:
        json_path = output.strip("path=")
        with open(json_path, "w", encoding="utf8", newline="") as output_file:
            output_file.write(ujson.dumps(response))



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import os
import re

from collections import namedtuple
from pathlib import PurePath
from typing import List

from polyaxon.logger import logger
from polyaxon.managers.base import BaseConfigManager
from polyaxon.utils import cli_constants
from polyaxon.utils.list_utils import to_list
from polyaxon.utils.path_utils import unix_style_path


class Pattern(namedtuple("Pattern", "pattern is_exclude re")):
    @staticmethod
    def create(pattern):
        if pattern[0:1] == "!":
            is_exclude = False
            pattern = pattern[1:]
        else:
            if pattern[0:1] == "\\":
                pattern = pattern[1:]
            is_exclude = True
        return Pattern(
            pattern=pattern,
            is_exclude=is_exclude,
            re=re.compile(translate(pattern), re.IGNORECASE),
        )

    def match(self, path):
        return bool(self.re.match(path))


def translate(pat):
    def _translate_segment():
        # pylint:disable=undefined-loop-variable
        if segment == "*":
            return "[^/]+"
        res = ""
        i, n = 0, len(segment)
        while i < n:
            c = segment[i : i + 1]
            i = i + 1
            if c == "*":
                res += "[^/]*"
            elif c == "?":
                res += "[^/]"
            elif c == "[":
                j = i
                if j < n and segment[j : j + 1] == "!":
                    j = j + 1
                if j < n and segment[j : j + 1] == "]":
                    j = j + 1
                while j < n and segment[j : j + 1] != "]":
                    j = j + 1
                if j >= n:
                    res += "\\["
                else:
                    stuff = segment[i:j].replace("\\", "\\\\")
                    i = j + 1
                    if stuff.startswith("!"):
                        stuff = "^" + stuff[1:]
                    elif stuff.startswith("^"):
                        stuff = "\\" + stuff
                    res += "[" + stuff + "]"
            else:
                res += re.escape(c)
        return res

    res = "(?ms)"

    if pat.startswith("**/"):
        pat = pat[2:]
        res += "(.*/)?"

    if pat.startswith("/"):
        pat = pat[1:]
    else:
        res += "(.*/)?"

    for i, segment in enumerate(pat.split("/")):
        if segment == "**":
            res += "(/.*)?"
            continue
        else:
            res += (re.escape("/") if i > 0 else "") + _translate_segment()

    if not pat.endswith("/"):
        res += "/?"

    return res + "\\Z"


class IgnoreConfigManager(BaseConfigManager):
    """Manages .polyaxonignore file in the current directory"""

    VISIBILITY = BaseConfigManager.VISIBILITY_LOCAL
    CONFIG_FILE_NAME = ".polyaxonignore"

    @staticmethod
    def _is_empty_or_comment(line):
        return not line or line.startswith("#")

    @staticmethod
    def _remove_trailing_spaces(line):
        """Remove trailing spaces unless they are quoted with a backslash."""
        while line.endswith(" ") and not line.endswith("\\ "):
            line = line[:-1]
        return line.replace("\\ ", " ")

    @classmethod
    def init_config(cls):
        cls.set_config(cli_constants.DEFAULT_IGNORE_LIST, init=True)

    @classmethod
    def find_matching(cls, path, patterns):
        """Yield all matching patterns for path."""
        for pattern in patterns:
            if pattern.match(path):
                yield pattern

    @classmethod
    def is_ignored(cls, path, patterns, is_dir: bool = False):
        """Check whether a path is ignored. For directories, include a trailing slash."""
        status = None
        path = "{}/".format(path.rstrip("/")) if is_dir else path
        for pattern in cls.find_matching(path, patterns):
            status = pattern.is_exclude
        return status

    @classmethod
    def read_file(cls, ignore_file):
        for line in ignore_file:
            line = line.rstrip("\r\n")

            if cls._is_empty_or_comment(line):
                continue

            yield cls._remove_trailing_spaces(line)

    @classmethod
    def get_patterns(cls, ignore_file):
        return [Pattern.create(line) for line in cls.read_file(ignore_file)]

    @staticmethod
    def get_push_patterns():
        return [
            Pattern.create("*.plx.json"),
            Pattern.create("*.plx.index"),
            Pattern.create("./.polyaxon"),
        ]

    @classmethod
    def get_config(cls):
        config_filepath = cls.get_config_filepath()

        if not os.path.isfile(config_filepath):
            # Return default patterns
            return cls.get_patterns(io.StringIO(cli_constants.DEFAULT_IGNORE_LIST))

        with open(config_filepath) as ignore_file:
            return cls.get_patterns(ignore_file)

    @classmethod
    def get_unignored_filepaths(
        cls, path: str = None, addtional_patterns: List[Pattern] = None
    ):
        config = to_list(cls.get_config(), check_none=True)
        config += to_list(addtional_patterns, check_none=True)
        unignored_files = []
        path = path or "."

        for root, dirs, files in os.walk(path):
            logger.debug("Root:%s, Dirs:%s", root, dirs)

            if cls.is_ignored(unix_style_path(root), config, is_dir=True):
                dirs[:] = []
                logger.debug("Ignoring directory : %s", root)
                continue

            for file_name in files:
                filepath = unix_style_path(os.path.join(root, file_name))
                if cls.is_ignored(filepath, config):
                    logger.debug("Ignoring file : %s", file_name)
                    continue

                unignored_files.append(os.path.join(root, file_name))

        return unignored_files

    @staticmethod
    def _matches_patterns(path, patterns):
        """Given a list of patterns, returns a if a path matches any pattern."""
        for glob in patterns:
            try:
                if PurePath(path).match(glob):
                    return True
            except TypeError:
                pass
        return False

    @classmethod
    def _ignore_path(cls, path, ignore_list=None, allowed_list=None):
        """Returns a whether a path should be ignored or not."""
        ignore_list = ignore_list or []
        allowed_list = allowed_list or []
        return cls._matches_patterns(path, ignore_list) and not cls._matches_patterns(
            path, allowed_list
        )

    @classmethod
    def get_value(cls, key):
        pass



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from marshmallow import fields

import polyaxon_sdk

from polyaxon.schemas.base import BaseCamelSchema
from polyaxon.schemas.fields.ref_or_obj import RefOrObject
from polyaxon.schemas.types.base import BaseTypeConfig


class TensorboardTypeSchema(BaseCamelSchema):
    port = RefOrObject(fields.Int(allow_none=True))
    uuids = RefOrObject(fields.List(fields.Str(allow_none=True)))
    use_names = RefOrObject(fields.Bool(allow_none=True))
    path_prefix = RefOrObject(fields.Str(allow_none=True))
    plugins = RefOrObject(fields.List(fields.Str(allow_none=True)))

    @staticmethod
    def schema_config():
        return V1TensorboardType


class V1TensorboardType(BaseTypeConfig, polyaxon_sdk.V1TensorboardType):
    """Tensorboard type.

    This type allows to initialize Tensorboard logs foe one or multiple operations.

    Args:
        port: int
        uuids: List[str]
        use_names: bool, optional
        path_prefix: str, optional
        plugins: List[str]

    ### YAML usage

    ### Usage in IO and params definition

    The inputs definition

    ```yaml
    >>> inputs:
    >>>   - name: tensorboard_content
    >>>     type: tensorboard
    ```

    The params usage

    ```yaml
    >>> params:
    >>>   tensorboard_content:
    >>>     value:
    >>>       port: 6006
    >>>       uuids: "d1410a914d18457589b91926d8c23db4,56f1a7f20f1d4f7f9e1a108b3c6b6031"
    >>>       useNames: true
    ```

    ### Usage in initializers

    ```yaml
     ```yaml
    >>> version:  1.1
    >>> kind: component
    >>> run:
    >>>   kind: service
    >>>   init:
    >>>   - tensorboard:
    >>>       uuids: "{{uuids}}"
    >>>       port: "{{globals.ports[0]}}"
    >>>       pathPrefix: "{{globals.base_url}}"
    >>>       useNames: true
    >>>       plugins: "tensorboard-plugin-profile"
    ```

    ### Python usage

    ### Usage in IO and params definition

    The inputs definition

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import types
    >>> from polyaxon.polyflow import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="tensorboard_content",
    >>>         type=types.TENSORBOARD,
    >>>     ),
    >>> ]
    ```

    The params usage

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import types
    >>> from polyaxon.polyflow import V1Param
    >>>
    >>> params = {
    >>>     "test1": V1Param(
    >>>         value=types.V1TensorboardType(
    >>>             port=6006,
    >>>             uuids="d1410a914d18457589b91926d8c23db4,56f1a7f20f1d4f7f9e1a108b3c6b6031",
    >>>             use_names=True,
    >>>         )
    >>>     ),
    >>> }
    ```

    ### Usage in initializers

    ```python
    >>> from polyaxon.polyflow import V1Component, V1Init, V1Job
    >>> from polyaxon.schemas.types import V1FileType
    >>> from polyaxon.k8s import k8s_schemas
    >>> component = V1Component(
    >>>     run=V1Job(
    >>>        init=[
    >>>             V1Init(
    >>>                 file=V1TensorboardType(
    >>>                     uuids="{{uuids}}",
    >>>                     port="{{globals.ports[0]}}",
    >>>                     path_prefix="{{globals.base_url}}",
    >>>                     use_names=True,
    >>>                     plugins="tensorboard-plugin-profile",
    >>>                 )
    >>>             ),
    >>>        ],
    >>>        container=k8s_schemas.V1Container(...)
    >>>     )
    >>> )
    ```

    ### Fields
      * port: port to expose the tensorboard service.
      * uuids: comma separated list of operation's uuids to load the tensorboard logs from.
      * useNames: an optional flag to initialize the paths under the operation's names.
      * pathPrefix: an optional path prefix to use for exposing the service.
      * plugins: an optional comma separated list of plugins to install before starting the tensorboard service.
    """

    IDENTIFIER = "tensorboard"
    SCHEMA = TensorboardTypeSchema
    REDUCED_ATTRIBUTES = [
        "port",
        "uuids",
        "useNames",
        "pathPrefix",
        "plugins",
    ]



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jinja2
import os

from typing import List, Tuple, Union

from polyaxon.builds.generator.dockerfile import POLYAXON_DOCKER_TEMPLATE
from polyaxon.utils.list_utils import to_list


class DockerFileGenerator:
    def __init__(self, build_context, destination="."):
        self.build_context = build_context
        self.dockerfile_path = os.path.join(destination, build_context.filename)

    def clean(self):
        # Clean dockerfile
        if not os.path.exists(self.dockerfile_path):
            return
        os.remove(self.dockerfile_path)

    def create(self, save=True):
        # Create DockerFile
        rendered_dockerfile = self.render()
        if save:
            with open(self.dockerfile_path, "w") as dockerfile:
                dockerfile.write(rendered_dockerfile)
        return rendered_dockerfile

    @staticmethod
    def get_copy(
        copy: Union[List[str], List[Tuple[str, str]]], workdir: str
    ) -> List[List[str]]:
        copy = to_list(copy, check_none=True)

        def handle_copy(c: Union[str, List]):
            if isinstance(c, str):
                return [c, workdir]
            return c

        return [handle_copy(c) for c in copy]

    def render(self):
        docker_template = jinja2.Template(POLYAXON_DOCKER_TEMPLATE)
        return docker_template.render(
            image=self.build_context.image,
            copy=self.get_copy(self.build_context.copy, self.build_context.workdir),
            post_run_copy=self.get_copy(
                self.build_context.post_run_copy, self.build_context.workdir
            ),
            run=to_list(self.build_context.run, check_none=True),
            env=to_list(self.build_context.env, check_none=True, check_dict=True),
            workdir=self.build_context.workdir,
            path=to_list(self.build_context.path, check_none=True),
            workdir_path=self.build_context.workdir_path,
            lang_env=self.build_context.lang_env,
            uid=self.build_context.uid,
            gid=self.build_context.gid,
            username=self.build_context.username or "polyaxon",
            shell=self.build_context.shell,
        )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from marshmallow import fields

import polyaxon_sdk

from polyaxon.connections.schemas import K8sResourceSchema
from polyaxon.schemas.base import BaseCamelSchema
from polyaxon.schemas.types.base import BaseTypeConfig


class K8sResourceTypeSchema(BaseCamelSchema):
    name = fields.Str(required=True)
    schema = fields.Nested(K8sResourceSchema, allow_none=True)
    is_requested = fields.Bool(allow_none=True)

    @staticmethod
    def schema_config():
        return V1K8sResourceType


class V1K8sResourceType(BaseTypeConfig, polyaxon_sdk.V1K8sResourceType):
    IDENTIFIER = "secret_resource"
    SCHEMA = K8sResourceTypeSchema
    REDUCED_ATTRIBUTES = ["name", "schema", "isRequested"]

    @classmethod
    def from_model(cls, model, is_requested=False) -> "V1K8sResourceType":
        schema = model.schema
        if hasattr(schema, "to_dict"):
            schema = schema.to_dict()
        return V1K8sResourceType.from_dict(
            {"name": model.name, "schema": schema, "isRequested": is_requested}
        )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

from polyaxon.contexts import paths as ctx_paths
from polyaxon.k8s import k8s_schemas
from polyaxon.polypod.common import constants
from polyaxon.schemas.types import V1ConnectionType, V1K8sResourceType


def get_mount_from_store(
    store: V1ConnectionType,
) -> Optional[k8s_schemas.V1VolumeMount]:
    if not store or not store.is_mount:
        return None

    return k8s_schemas.V1VolumeMount(
        name=store.name,
        mount_path=store.schema.mount_path,
        read_only=store.schema.read_only,
    )


def get_mount_from_resource(
    resource: V1K8sResourceType,
) -> Optional[k8s_schemas.V1VolumeMount]:
    if not resource or not resource.schema.mount_path:
        return None

    return k8s_schemas.V1VolumeMount(
        name=resource.name, mount_path=resource.schema.mount_path, read_only=True
    )


def get_docker_context_mount() -> k8s_schemas.V1VolumeMount:
    return k8s_schemas.V1VolumeMount(
        name=constants.VOLUME_MOUNT_DOCKER, mount_path=ctx_paths.CONTEXT_MOUNT_DOCKER
    )


def get_auth_context_mount(read_only=None) -> k8s_schemas.V1VolumeMount:
    return k8s_schemas.V1VolumeMount(
        name=constants.VOLUME_MOUNT_CONFIGS,
        mount_path=ctx_paths.CONTEXT_MOUNT_CONFIGS,
        read_only=read_only,
    )


def get_artifacts_context_mount(read_only=None) -> k8s_schemas.V1VolumeMount:
    return k8s_schemas.V1VolumeMount(
        name=constants.VOLUME_MOUNT_ARTIFACTS,
        mount_path=ctx_paths.CONTEXT_MOUNT_ARTIFACTS,
        read_only=read_only,
    )


def get_connections_context_mount(
    name: str, mount_path: str
) -> k8s_schemas.V1VolumeMount:
    return k8s_schemas.V1VolumeMount(name=name, mount_path=mount_path)


def get_shm_context_mount() -> k8s_schemas.V1VolumeMount:
    """
    Mount an tmpfs volume to /dev/shm.
    This will set /dev/shm size to half of the RAM of node.
    By default, /dev/shm is very small, only 64MB.
    Some experiments will fail due to lack of share memory,
    such as some experiments running on Pytorch.
    """
    return k8s_schemas.V1VolumeMount(
        name=constants.VOLUME_MOUNT_SHM, mount_path=ctx_paths.CONTEXT_MOUNT_SHM
    )


def get_mounts(
    use_auth_context: bool,
    use_docker_context: bool,
    use_shm_context: bool,
    use_artifacts_context: bool,
) -> List[k8s_schemas.V1VolumeMount]:
    mounts = []
    if use_auth_context:
        mounts.append(get_auth_context_mount(read_only=True))
    if use_artifacts_context:
        mounts.append(get_artifacts_context_mount(read_only=False))
    if use_docker_context:
        mounts.append(get_docker_context_mount())
    if use_shm_context:
        mounts.append(get_shm_context_mount())

    return mounts



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sys
import urllib.parse as urlparse
import yaml

from typing import Dict, List, Tuple
from urllib.parse import parse_qs

import click

from rich import box
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from rich.syntax import Syntax
from rich.table import Column, Table
from rich.theme import Theme

from polyaxon.schemas.api.resources import ContainerResourcesConfig
from polyaxon.utils.humanize import humanize_timesince
from polyaxon.utils.list_utils import to_list
from traceml.processors.units_processors import to_percentage, to_unit_memory


def get_meta_response(response):
    def get_pagination(url):
        parsed = urlparse.urlparse(url)
        parsed_query = parse_qs(parsed.query)
        limit = parsed_query.get("limit")
        offset = parsed_query.get("offset")
        res = []
        if limit is not None:
            res.append("--limit={}".format(limit[0]))
        if offset is not None:
            res.append("--offset={}".format(offset[0]))
        return " ".join(res)

    results = {}
    if response.next:
        try:
            results["next"] = get_pagination(response.next)
        except Exception:  # noqa
            results["next"] = response.next
    if response.previous:
        try:
            results["previous"] = get_pagination(response.previous)
        except Exception:  # noqa
            results["previous"] = response.previous
    if response.count:
        results["count"] = response.count
    return results


def humanize_attrs(key, value, rounding=2):
    if key in [
        "created_at",
        "updated_at",
        "started_at",
        "finished_at",
        "schedule_at",
        "last_update_time",
        "last_transition_time",
    ]:
        return humanize_timesince(value)
    if key in ["cpu_percentage"]:
        return to_percentage(value, rounding)
    if key in ["memory_free", "memory_used", "memory_total"]:
        return to_unit_memory(value)
    if isinstance(value, str):
        return value
    if value is None:
        return ""
    return json.dumps(value)


def list_dicts_to_tabulate(
    list_dicts,
    exclude_attrs=None,
    include_attrs=None,
    humanize_values=True,
    upper_keys: bool = True,
):
    exclude_attrs = exclude_attrs or {}
    results = []
    if include_attrs:  # If include_attrs disable exclude_attrs
        exclude_attrs = {}
    for d_value in list_dicts:
        r_value = {}
        for k, v in d_value.items():
            if k in exclude_attrs:
                continue
            if include_attrs and k not in include_attrs:
                continue

            if humanize_values:
                v = humanize_attrs(k, v)

            if upper_keys:
                k = k.upper()
            r_value[k] = v
        results.append(r_value)
    return results


def list_dicts_to_csv(
    list_dicts,
    exclude_attrs=None,
    include_attrs=None,
):
    exclude_attrs = exclude_attrs or {}
    results = []
    if include_attrs:  # If include_attrs disable exclude_attrs
        exclude_attrs = {}
    for d_value in list_dicts:
        result = {}
        for k, v in d_value.items():
            if k in exclude_attrs:
                continue
            if include_attrs and k not in include_attrs:
                continue
            result[k] = v
        results.append(result)

    return results


def dict_to_tabulate(d_value, exclude_attrs=None, humanize_values=True):
    exclude_attrs = exclude_attrs or {}
    results = {}
    for k, v in d_value.items():
        if k in exclude_attrs:
            continue

        if humanize_values:
            v = humanize_attrs(k, v)

        results[k.upper()] = v

    return results


def dict_tabulate(dict_value, is_list_dict=False):
    if is_list_dict:
        headers = dict_value[0].keys() if dict_value else []
        table = Printer.get_table(*headers)
        for d in dict_value:
            table.add_row(*d.values())
        Printer.print(table)
    else:
        table = Printer.get_table(show_header=False, padding=0, box=box.SIMPLE)
        for k, v in dict_value.items():
            table.add_row(k, humanize_attrs(k, v))
        Printer.print(table)


class Printer:
    COLORS = ["yellow", "blue", "magenta", "green", "cyan", "red", "white"]
    console = Console(
        theme=Theme(
            {
                "header": "yellow",
                "success": "green",
                "info": "cyan",
                "warning": "magenta",
                "error": "red",
                "white": "white",
            }
        ),
        markup=True,
    )

    @staticmethod
    def get_progress():
        return Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            DownloadColumn(),
            TransferSpeedColumn(),
            TaskProgressColumn(),
            TextColumn("eta"),
            TimeRemainingColumn(),
            TextColumn("elapsed"),
            TimeElapsedColumn(),
        )

    @classmethod
    def get_live(cls):
        return Live(console=cls.console)

    @staticmethod
    def get_table(*args, **kwargs):
        return Table(*[Column(header=h, no_wrap=True) for h in args], **kwargs)

    @staticmethod
    def pprint(value):
        """Prints as formatted JSON"""
        click.echo(json.dumps(value, sort_keys=True, indent=4, separators=(",", ": ")))

    @classmethod
    def print_md(cls, md: str):
        cls.console.print(Markdown(md))

    @classmethod
    def print_text(cls, value: str):
        syntax = Syntax(value, "txt", theme="dracula", line_numbers=False)
        cls.console.print(syntax)

    @classmethod
    def print_yaml(cls, value: any):
        if isinstance(value, str):
            value = yaml.safe_load(value)
        value = yaml.safe_dump(value, sort_keys=True, indent=2)
        syntax = Syntax(value, "yaml", theme="dracula", line_numbers=False)
        cls.console.print(syntax)

    @classmethod
    def print_json(cls, value: any):
        if isinstance(value, str):
            value = json.loads(value)
        value = json.dumps(value, sort_keys=True, indent=4, separators=(",", ": "))
        syntax = Syntax(value, "json", theme="dracula", line_numbers=False)
        cls.console.print(syntax)

    @classmethod
    def print_help(cls, command_help: str = None):
        if command_help:
            cls.console.print(
                "Please run [white]`polyaxon {} --help`[/white] for more details".format(
                    command_help
                ),
                style="info",
            )

    @classmethod
    def print(cls, text):
        cls.console.print(text)

    @classmethod
    def print_heading(cls, text):
        cls.print_header("\n{}\n".format(text))

    @classmethod
    def print_header(cls, text):
        cls.console.print(text, style="header")

    @classmethod
    def print_warning(cls, text, command_help: str = None):
        cls.console.print(text, style="warning")
        if command_help:
            cls.print_help(command_help)

    @classmethod
    def print_success(cls, text):
        cls.console.print(text, style="success")

    @classmethod
    def print_error(cls, text, sys_exit: bool = False, command_help: str = None):
        cls.console.print(text, style="error")
        if command_help:
            cls.print_help(command_help)
        if sys_exit:
            sys.exit(1)

    @classmethod
    def print_tip(cls, text):
        cls.console.print(text, style="white")

    @classmethod
    def print_info(cls, text):
        cls.console.print(text, style="info")

    @staticmethod
    def add_log_color(value, color):
        return click.style("{}".format(value), fg=color)

    @classmethod
    def add_color(cls, value, style):
        return "[{style}]{value}[/{style}]".format(value=value, style=style)

    @classmethod
    def get_colored_status(cls, status):
        if status == "created":
            return cls.add_color(status, "info")
        elif status == "succeeded":
            return cls.add_color(status, style="success")
        elif status in ["failed", "stopped", "upstream_failed"]:
            return cls.add_color(status, style="error")
        elif status == "done":
            return cls.add_color(status, style="white")

        return cls.add_color(status, style="header")

    @classmethod
    def add_status_color(cls, obj_dict, status_key="status"):
        if obj_dict.get(status_key) is None:
            return obj_dict

        obj_dict[status_key] = cls.get_colored_status(obj_dict[status_key])
        return obj_dict

    @classmethod
    def add_memory_unit(cls, obj_dict, keys):
        keys = to_list(keys)
        for key in keys:
            obj_dict[key] = to_unit_memory(obj_dict[key])
        return obj_dict

    @classmethod
    def decorate_format_value(cls, text_format, values, color):
        values = to_list(values)
        values = [cls.add_color(value, color) for value in values]
        click.echo(text_format.format(*values))

    @staticmethod
    def log(value, nl=False):
        click.echo(value, nl=nl)

    @classmethod
    def resources(cls, jobs_resources):
        jobs_resources = to_list(jobs_resources)
        click.clear()
        table = Printer.get_table("Job", "Mem Usage / Total", "CPU% - CPUs")
        for job_resources in jobs_resources:
            job_resources = ContainerResourcesConfig.from_dict(job_resources)
            line = [
                job_resources.job_name,
                "{} / {}".format(
                    to_unit_memory(job_resources.memory_used),
                    to_unit_memory(job_resources.memory_limit),
                ),
                "{} - {}".format(
                    to_percentage(job_resources.cpu_percentage / 100),
                    job_resources.n_cpus,
                ),
            ]
            table.add_row(*line)
        Printer.print(table)
        sys.stdout.flush()

    @classmethod
    def gpu_resources(cls, jobs_resources):
        jobs_resources = to_list(jobs_resources)
        click.clear()
        table = Printer.get_table(
            "job_name",
            "name",
            "GPU Usage",
            "GPU Mem Usage / Total",
            "GPU Temperature",
            "Power Draw / Limit",
        )
        non_gpu_jobs = 0
        for job_resources in jobs_resources:
            job_resources = ContainerResourcesConfig.from_dict(job_resources)
            line = []
            if not job_resources.gpu_resources:
                non_gpu_jobs += 1
                continue
            for gpu_resources in job_resources.gpu_resources:
                line += [
                    job_resources.job_name,
                    gpu_resources.name,
                    to_percentage(gpu_resources.utilization_gpu / 100),
                    "{} / {}".format(
                        to_unit_memory(gpu_resources.memory_used),
                        to_unit_memory(gpu_resources.memory_total),
                    ),
                    gpu_resources.temperature_gpu,
                    "{} / {}".format(
                        gpu_resources.power_draw, gpu_resources.power_limit
                    ),
                ]
            table.add_row(*line)
        if non_gpu_jobs == len(jobs_resources):
            Printer.print_error(
                "No GPU job was found, please run `resources` command without `-g | --gpu` option."
            )
            exit(1)
        Printer.print(table)
        sys.stdout.flush()


def flatten_keys(
    objects: List[Dict], columns: List[str], columns_prefix: Dict = None
) -> Tuple[List[Dict], Dict]:
    # Extend run with params_keys
    keys = set([])
    columns_prefix = columns_prefix or {}
    prefixed_columns = {}

    def process_objects():
        results = {}
        for k, v in col_values.items():
            results["{}.{}".format(column_prefix, k)] = v
            prefixed_columns[k] = "{}.{}".format(column_prefix, k)
        return results

    for col in columns:
        for obj in objects:
            col_values = obj.pop(col, {}) or {}
            if col in columns_prefix:
                column_prefix = columns_prefix[col]
                col_values = process_objects()
            col_keys = col_values.keys()
            keys |= set(col_keys)
            obj.update(col_values)

    # Check that all obj have all metrics
    # TODO: optimize this process
    for obj in objects:
        obj_keys = set(obj.keys())
        for key in keys:
            if key not in obj_keys:
                obj[key] = None

    return objects, prefixed_columns



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools

from urllib3.exceptions import HTTPError

from polyaxon import settings
from polyaxon.client import PolyaxonClient
from polyaxon.logger import logger
from polyaxon_sdk.rest import ApiException


def client_handler(
    check_no_op: bool = True,
    check_offline: bool = False,
    can_log_events: bool = False,
    can_log_outputs: bool = False,
):
    """
    The `ClientHandlerDecorator` is a decorator to handle several checks in PolyaxonClient.

     * check_offline: to ignore any decorated function when POLYAXON_IS_OFFLINE env var is found.
     * check_no_op: to ignore any decorated function when NO_OP env var is found.
     * handle_openapi_exceptions: to handle exception of OpenApi and generate better
        debugging outputs.
     * can_log_events: to check if there's an event logger instance on the object.
     * can_log_outputs: to check if there's an outputs path set on the run.
     * openapi_extra_context: to augment openapi errors.

    usage example with class method:
        @client_handler(check_no_op=True)
        def my_func(self, *args, **kwargs):
            ...
            return ...

    usage example with a function:
        @client_handler(check_no_op=True)
        def my_func(arg1, arg2):
            ...
            return ...
    """

    def _check_global_or_inline_config(args, config_key):
        self_arg = args[0] if args else None
        config_value = getattr(self_arg, f"_{config_key}", None)
        return get_global_or_inline_config(
            config_key=config_key,
            config_value=config_value,
            client=getattr(self_arg, "_client", None),
        )

    def client_handler_wrapper(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if check_no_op and _check_global_or_inline_config(args, "no_op"):
                logger.debug("Using NO_OP mode")
                return None
            if check_offline and _check_global_or_inline_config(args, "is_offline"):
                logger.debug("Using IS_OFFLINE mode")
                return None
            if args:
                self_arg = args[0]

                if can_log_events and (
                    not hasattr(self_arg, "_event_logger")
                    or self_arg._event_logger is None  # pylint:disable=protected-access
                ):
                    logger.warning(
                        "You should set an event logger before calling: {}".format(
                            f.__name__
                        )
                    )

                if can_log_outputs and (
                    not hasattr(self_arg, "_outputs_path")
                    or self_arg._outputs_path is None  # pylint:disable=protected-access
                ):
                    logger.warning(
                        "You should set an an outputs path before calling: {}".format(
                            f.__name__
                        )
                    )

            try:
                return f(*args, **kwargs)
            except (ApiException, HTTPError) as e:
                logger.debug("Client config:\n%s\n", settings.CLIENT_CONFIG.to_dict())
                raise e

        return wrapper

    return client_handler_wrapper


def get_global_or_inline_config(
    config_key: str,
    config_value: bool = None,
    client: PolyaxonClient = None,
):
    if config_value is not None:
        return config_value
    if client and client.config and getattr(client.config, config_key) is not None:
        return getattr(client.config, config_key)
    return getattr(settings.CLIENT_CONFIG, config_key, None)



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon import settings
from polyaxon.proxies.schemas.base import get_config

OPTIONS = r"""
#!/bin/bash
set -e
set -o pipefail

{cmd}
"""


def get_forward_cmd():
    if not settings.PROXIES_CONFIG.has_forward_proxy:
        return

    cmd = None
    if settings.PROXIES_CONFIG.forward_proxy_kind == "transparent":
        cmd = "socat TCP4-LISTEN:8443,reuseaddr,fork TCP:{proxy_host}:{proxy_port}".format(
            proxy_host=settings.PROXIES_CONFIG.forward_proxy_host,
            proxy_port=settings.PROXIES_CONFIG.forward_proxy_port,
        )
    elif (
        settings.PROXIES_CONFIG.forward_proxy_kind is None
        or settings.PROXIES_CONFIG.forward_proxy_kind == "connect"
    ):
        cmd = (
            "socat TCP4-LISTEN:8443,reuseaddr,fork,bind=127.0.0.1 "
            "PROXY:{proxy_host}:{api_host}:{api_port},proxyport={proxy_port}".format(
                api_host=settings.PROXIES_CONFIG.api_host,
                api_port=settings.PROXIES_CONFIG.api_port,
                proxy_host=settings.PROXIES_CONFIG.forward_proxy_host,
                proxy_port=settings.PROXIES_CONFIG.forward_proxy_port,
            )
        )

    if not cmd:
        return

    return get_config(
        options=OPTIONS,
        indent=0,
        cmd=cmd,
    )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from marshmallow import fields

from polyaxon.schemas.base import BaseCamelSchema, BaseConfig


class EmailSchema(BaseCamelSchema):
    enabled = fields.Bool(allow_none=True)
    email_from = fields.Str(allow_none=True, data_key="from")
    host = fields.Str(allow_none=True)
    port = fields.Int(allow_none=True)
    use_tls = fields.Bool(allow_none=True)
    host_user = fields.Str(allow_none=True)
    host_password = fields.Str(allow_none=True)
    backend = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return EmailConfig


class EmailConfig(BaseConfig):
    SCHEMA = EmailSchema
    REDUCED_ATTRIBUTES = [
        "enabled",
        "from",
        "host",
        "port",
        "useTls",
        "hostUser",
        "hostPassword",
        "backend",
    ]

    def __init__(
        self,  # noqa
        enabled=None,
        email_from=None,
        host=None,
        port=None,
        use_tls=None,
        host_user=None,
        host_password=None,
        backend=None,
    ):
        self.enabled = enabled
        self.email_from = email_from
        self.host = host
        self.port = port
        self.use_tls = use_tls
        self.host_user = host_user
        self.host_password = host_password
        self.backend = backend



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from polyaxon.config_reader.manager import ConfigManager
from polyaxon.config_reader.spec import ConfigSpec
from polyaxon.contexts import paths as ctx_paths
from polyaxon.managers.base import BaseConfigManager
from polyaxon.schemas.api.authentication import AccessTokenConfig


class AuthConfigManager(BaseConfigManager):
    """Manages access token configuration .auth file."""

    VISIBILITY = BaseConfigManager.VISIBILITY_GLOBAL
    CONFIG_FILE_NAME = ".auth"
    CONFIG = AccessTokenConfig

    @classmethod
    def get_config_from_env(cls) -> AccessTokenConfig:
        tmp_path = os.path.join(
            ctx_paths.CONTEXT_TMP_POLYAXON_PATH, cls.CONFIG_FILE_NAME
        )
        user_path = os.path.join(
            ctx_paths.CONTEXT_USER_POLYAXON_PATH, cls.CONFIG_FILE_NAME
        )
        auth_config = ConfigManager.read_configs(
            [
                os.environ,
                ConfigSpec(tmp_path, config_type=".json", check_if_exists=False),
                ConfigSpec(user_path, config_type=".json", check_if_exists=False),
                ConfigSpec(
                    ctx_paths.CONTEXT_MOUNT_AUTH,
                    config_type=".json",
                    check_if_exists=False,
                ),
                {"dummy": "dummy"},
            ]
        )
        return AccessTokenConfig.from_dict(auth_config.data)



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, Optional

from polyaxon.polyflow import V1CompiledOperation
from polyaxon.polypod.compiler import converter, resolver


def make(
    owner_name: str,
    project_name: str,
    project_uuid: str,
    run_name: str,
    run_uuid: str,
    run_path: str,
    compiled_operation: V1CompiledOperation,
    params: Optional[Dict],
    default_sa: str = None,
    internal_auth: bool = False,
    default_auth: bool = False,
    converters: Dict[str, Any] = converter.CORE_CONVERTERS,
):
    resolver_obj, compiled_operation = resolver.resolve(
        compiled_operation=compiled_operation,
        owner_name=owner_name,
        project_name=project_name,
        project_uuid=project_uuid,
        run_name=run_name,
        run_path=run_path,
        run_uuid=run_uuid,
        params=params,
    )
    return converter.convert(
        namespace=resolver_obj.namespace,
        owner_name=resolver_obj.owner_name,
        project_name=resolver_obj.project_name,
        run_name=resolver_obj.run_name,
        run_path=resolver_obj.run_path,
        run_uuid=resolver_obj.run_uuid,
        compiled_operation=compiled_operation,
        connection_by_names=resolver_obj.connection_by_names,
        internal_auth=internal_auth,
        artifacts_store=resolver_obj.artifacts_store,
        secrets=resolver_obj.secrets,
        config_maps=resolver_obj.config_maps,
        polyaxon_sidecar=resolver_obj.polyaxon_sidecar,
        polyaxon_init=resolver_obj.polyaxon_init,
        default_sa=default_sa,
        converters=converters,
        default_auth=default_auth,
    )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy

from collections.abc import Mapping
from typing import Dict, List, Union

from polyaxon import pkg
from polyaxon.env_vars.getters.queue import get_queue_info
from polyaxon.exceptions import PolyaxonfileError
from polyaxon.polyaxonfile.specs import (
    CompiledOperationSpecification,
    OperationSpecification,
    get_specification,
    kinds,
)
from polyaxon.polyflow import V1Component, V1Init, V1Matrix, V1MatrixKind, V1Operation
from polyaxon.schemas import V1PatchStrategy
from polyaxon.utils.bool_utils import to_bool


def get_op_specification(
    config: Union[V1Component, V1Operation] = None,
    hub: str = None,
    params: Dict = None,
    hparams: Dict = None,
    matrix_kind: str = None,
    matrix_concurrency: int = None,
    matrix_num_runs: int = None,
    matrix: V1Matrix = None,
    presets: List[str] = None,
    queue: str = None,
    nocache: bool = None,
    cache: Union[int, str, bool] = None,
    approved: Union[int, str, bool] = None,
    validate_params: bool = True,
    preset_files: List[str] = None,
    git_init: V1Init = None,
) -> V1Operation:
    if cache and nocache:
        raise PolyaxonfileError("Received both 'cache' and 'nocache'")
    op_data = {
        "version": config.version if config else pkg.SCHEMA_VERSION,
        "kind": kinds.OPERATION,
    }
    if params:
        if not isinstance(params, Mapping):
            raise PolyaxonfileError(
                "Params: `{}` must be a valid mapping".format(params)
            )
        op_data["params"] = params
    if hparams:
        if not isinstance(hparams, Mapping):
            raise PolyaxonfileError(
                "Hyper-Params: `{}` must be a valid mapping".format(hparams)
            )
        op_data["matrix"] = {
            "kind": matrix_kind or V1MatrixKind.GRID,
            "concurrency": matrix_concurrency or 1,
            "params": hparams,
        }
        if matrix_num_runs:
            op_data["matrix"]["numRuns"] = matrix_num_runs
    if matrix:
        op_data["matrix"] = (
            matrix if isinstance(matrix, Mapping) else matrix.to_light_dict()
        )
    if presets:
        op_data["presets"] = presets
    if queue:
        # Check only
        get_queue_info(queue)
        op_data["queue"] = queue
    if cache is not None:
        op_data["cache"] = {"disable": not to_bool(cache)}
    if nocache:
        op_data["cache"] = {"disable": True}
    # Handle approval logic
    if approved is not None:
        op_data["isApproved"] = to_bool(approved)

    if config and config.kind == kinds.COMPONENT:
        op_data["component"] = config.to_dict()
        config = get_specification(data=[op_data])
    elif config and config.kind == kinds.OPERATION:
        config = get_specification(data=[config.to_dict(), op_data])
    elif hub:
        op_data["hubRef"] = hub
        config = get_specification(data=[op_data])

    if hub and config.hub_ref is None:
        config.hub_ref = hub

    # Check if there's presets
    for preset_plx_file in preset_files:
        preset_plx_file = OperationSpecification.read(preset_plx_file, is_preset=True)
        config = config.patch(preset_plx_file, strategy=preset_plx_file.patch_strategy)
    # Turn git_init to a pre_merge preset
    if git_init:
        git_preset = V1Operation(
            run_patch={"init": [git_init.to_dict()]}, is_preset=True
        )
        config = config.patch(git_preset, strategy=V1PatchStrategy.PRE_MERGE)

    # Sanity check if params were passed and we are not dealing with a hub component
    params = copy.deepcopy(config.params)
    if validate_params:
        # Avoid in-place patch
        run_config = get_specification(config.to_dict())
        run_config = OperationSpecification.compile_operation(run_config)
        run_config.validate_params(params=params, is_template=False)
        if run_config.is_dag_run:
            CompiledOperationSpecification.apply_operation_contexts(run_config)
    return config



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from typing import Callable, List, Optional, Union

import click

from urllib3.exceptions import HTTPError

from polyaxon.cli.dashboard import get_dashboard_url
from polyaxon.cli.errors import handle_cli_error
from polyaxon.client import PolyaxonClient, ProjectClient
from polyaxon.lifecycle import V1ProjectVersionKind
from polyaxon.utils.formatting import (
    Printer,
    dict_tabulate,
    dict_to_tabulate,
    get_meta_response,
    list_dicts_to_tabulate,
)
from polyaxon.utils.fqn_utils import get_versioned_entity_full_name
from polyaxon.utils.query_params import get_query_params
from polyaxon.utils.validation import validate_tags
from polyaxon_sdk.rest import ApiException


def get_version_details(response, content_callback: Callable = None):
    content = response.content
    meta_info = response.meta_info
    readme = response.readme
    response = dict_to_tabulate(
        response.to_dict(),
        humanize_values=True,
        exclude_attrs=["content", "meta_info", "stage_conditions", "readme"],
    )

    Printer.print_heading("Version info:")
    dict_tabulate(response)

    if meta_info:
        artifacts = meta_info.pop("artifacts", None)
        lineage = meta_info.pop("lineage", artifacts)
        if meta_info:
            Printer.print_heading("Version meta info:")
            dict_tabulate(meta_info)

        if lineage:
            Printer.print_heading("Version artifacts lineage:")
            Printer.print_json(lineage)

    if readme:
        Printer.print_heading("Version readme:")
        Printer.print_md(readme)

    def get_content(_content: str):
        if _content:
            Printer.print_heading("Content:")
            Printer.print_yaml(_content)

    content_callback = content_callback or get_content
    content_callback(content)


def get_version_stages_table(stage, conditions):
    table = Printer.get_table()

    if not conditions:
        Printer.print(table)
        return

    Printer.print(
        "Latest stage: {}".format(
            Printer.add_status_color({"stage": stage}, status_key="stage")["stage"]
        )
    )
    objects = [
        dict_to_tabulate(Printer.add_status_color(o.to_dict(), status_key="type"))
        for o in conditions
    ]

    if not objects:
        Printer.print(table)
        return

    if not table.columns:
        for c in objects[0].keys():
            table.add_column(c)
    if objects:
        for o in objects:
            table.add_row(*o.values())

    Printer.print(table)


def list_project_versions_response(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    limit: str = None,
    offset: str = None,
    query: str = None,
    sort: str = None,
    client: PolyaxonClient = None,
):
    polyaxon_client = ProjectClient(owner=owner, project=project_name, client=client)
    params = get_query_params(limit=limit, offset=offset, query=query, sort=sort)
    try:
        return polyaxon_client.list_versions(kind=kind, **params)
    except (ApiException, HTTPError) as e:
        message = "Could not get list of {} versions for {}/{}.".format(
            kind, owner, project_name
        )
        handle_cli_error(e, message=message)
        sys.exit(1)


def list_project_versions(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    limit: str = None,
    offset: str = None,
    query: str = None,
    sort: str = None,
    client: PolyaxonClient = None,
):
    version_info = "<owner: {}> <project: {}>".format(owner, project_name)
    response = list_project_versions_response(
        owner=owner,
        project_name=project_name,
        kind=kind,
        limit=limit,
        offset=offset,
        query=query,
        sort=sort,
        client=client,
    )
    meta = get_meta_response(response)
    if meta:
        Printer.print_heading("Versions for {}".format(version_info))
        Printer.print_heading("Navigation:")
        dict_tabulate(meta)
    else:
        Printer.print_header("No version found for {}".format(version_info))

    objects = list_dicts_to_tabulate(
        [o.to_dict() for o in response.results],
        humanize_values=True,
        exclude_attrs=[
            "uuid",
            "readme",
            "description",
            "owner",
            "project",
            "role",
            "content",
            "connection",
            "meta_info",
            "run",
            "artifacts",
            "stage_conditions",
        ],
    )
    if objects:
        Printer.print_heading("Versions:")
        dict_tabulate(objects, is_list_dict=True)


def register_project_version(
    owner: str,
    project_name: str,
    version: str,
    kind: V1ProjectVersionKind,
    description: str = None,
    tags: Optional[Union[str, List[str]]] = None,
    content: str = None,
    run: str = None,
    connection: str = None,
    artifacts: Optional[Union[str, List[str]]] = None,
    force: bool = False,
):
    version = version or "latest"
    fqn_version = get_versioned_entity_full_name(owner, project_name, version)
    polyaxon_client = ProjectClient(owner=owner, project=project_name)

    try:
        _version = polyaxon_client.register_version(
            kind=kind,
            version=version,
            description=description,
            tags=tags,
            content=content,
            run=run,
            connection=connection,
            artifacts=artifacts,
            force=force,
        )
    except (ApiException, HTTPError) as e:
        handle_cli_error(
            e, message="Could not create version `{}`.".format(fqn_version)
        )
        sys.exit(1)

    Printer.print_success("Version `{}` was created successfully.".format(fqn_version))
    Printer.print(
        "You can view this version on Polyaxon UI: {}".format(
            get_dashboard_url(
                subpath="{}/{}/{}s/{}".format(owner, project_name, kind, version)
            )
        )
    )


def copy_project_version(
    owner: str,
    project_name: str,
    version: str,
    kind: V1ProjectVersionKind,
    to_project: str = None,
    name: str = None,
    description: str = None,
    tags: Optional[Union[str, List[str]]] = None,
    content: str = None,
    force: bool = False,
):
    version = version or "latest"
    fqn_version = get_versioned_entity_full_name(owner, project_name, version)
    polyaxon_client = ProjectClient(owner=owner, project=project_name)

    try:
        _version = polyaxon_client.copy_version(
            kind=kind,
            version=version,
            to_project=to_project,
            name=name,
            description=description,
            tags=tags,
            content=content,
            force=force,
        )
    except (ApiException, HTTPError) as e:
        handle_cli_error(e, message="Could not copy version `{}`.".format(fqn_version))
        sys.exit(1)

    fqn_copied_version = get_versioned_entity_full_name(
        owner,
        to_project or project_name,
        _version.name,
    )
    Printer.print_success(
        "Version `{}` was copied successfully to `{}`.".format(
            fqn_version, fqn_copied_version
        )
    )
    Printer.print(
        "You can view this version on Polyaxon UI: {}".format(
            get_dashboard_url(
                subpath="{}/{}/{}s/{}".format(
                    owner, to_project or project_name, kind, _version.name
                )
            )
        )
    )


def get_project_version(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    version: str,
    content_callback: Callable = None,
    client: PolyaxonClient = None,
):
    fqn_version = get_versioned_entity_full_name(owner, project_name, version)
    polyaxon_client = ProjectClient(owner=owner, project=project_name, client=client)

    try:
        response = polyaxon_client.get_version(kind, version)
        get_version_details(response, content_callback)
    except (ApiException, HTTPError) as e:
        handle_cli_error(
            e,
            message="Could not get {} version `{}`.".format(
                kind,
                fqn_version,
            ),
            sys_exit=True,
        )


def get_project_version_stages(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    version: str,
    client: PolyaxonClient = None,
):
    fqn_version = get_versioned_entity_full_name(owner, project_name, version)
    polyaxon_client = ProjectClient(owner=owner, project=project_name, client=client)

    try:
        stage, stage_conditions = polyaxon_client.get_version_stages(kind, version)
        get_version_stages_table(stage, stage_conditions)
    except (ApiException, HTTPError) as e:
        handle_cli_error(
            e,
            message="Could not get {} version `{}`.".format(
                kind,
                fqn_version,
            ),
            sys_exit=True,
        )


def delete_project_version(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    version: str,
    yes: bool = False,
):
    fqn_version = get_versioned_entity_full_name(owner, project_name, version)
    polyaxon_client = ProjectClient(owner=owner, project=project_name)

    if not yes and not click.confirm(
        "Are sure you want to delete {} version `{}`".format(kind, fqn_version)
    ):
        Printer.print("Exiting without deleting {} version.".format(kind))
        sys.exit(1)

    try:
        polyaxon_client.delete_version(kind, version)
        Printer.print_success(
            "The {} version `{}` was delete successfully".format(kind, fqn_version)
        )
    except (ApiException, HTTPError) as e:
        handle_cli_error(
            e,
            message="Could not delete the {} version `{}`.".format(kind, fqn_version),
        )
        sys.exit(1)


def update_project_version(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    version: str,
    name: str = None,
    description: str = None,
    tags: Optional[Union[str, List[str]]] = None,
    content_callback: Callable = None,
):
    fqn_version = get_versioned_entity_full_name(owner, project_name, version)
    polyaxon_client = ProjectClient(owner=owner, project=project_name)

    update_dict = {}
    if name:
        update_dict["name"] = name

    if description:
        update_dict["description"] = description

    tags = validate_tags(tags, validate_yaml=True)
    if tags:
        update_dict["tags"] = tags

    if not update_dict:
        Printer.print_warning(
            "No argument was provided to update the {} version {}.".format(
                kind, fqn_version
            )
        )
        sys.exit(1)

    try:
        response = polyaxon_client.patch_version(kind, version, update_dict)
        Printer.print_success("The {} version updated.".format(kind))
        get_version_details(response, content_callback)
    except (ApiException, HTTPError) as e:
        handle_cli_error(
            e,
            message="Could not update the {} version `{}`.".format(kind, fqn_version),
        )
        sys.exit(1)


def transfer_project_version(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    version: str,
    to_project: str,
):
    fqn_version = get_versioned_entity_full_name(owner, project_name, version)
    polyaxon_client = ProjectClient(owner=owner, project=project_name)

    try:
        polyaxon_client.transfer_version(kind, version, to_project)
        Printer.print_success(
            "The `{}` version was transferred to `{}`.".format(kind, to_project)
        )
    except (ApiException, HTTPError) as e:
        handle_cli_error(
            e,
            message="Could not transfer the {} version `{}` to `{}`.".format(
                kind, fqn_version, to_project
            ),
        )
        sys.exit(1)


def stage_project_version(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    version: str,
    to: str,
    reason: str = None,
    message: str = None,
):
    fqn_version = get_versioned_entity_full_name(owner, project_name, version)
    polyaxon_client = ProjectClient(owner=owner, project=project_name)

    if not to:
        Printer.print_warning(
            "No argument was provided to update the version stage, "
            "please provide a correct `--to` value."
        )
        sys.exit(1)

    try:
        polyaxon_client.stage_version(
            kind, version, stage=to, reason=reason or "CliStageUpdate", message=message
        )
        Printer.print_success("The {} version's stage was updated.".format(kind))
    except (ApiException, HTTPError) as e:
        handle_cli_error(
            e,
            message="Could not update the stage for {} version `{}`.".format(
                kind, fqn_version
            ),
        )
        sys.exit(1)


def open_project_version_dashboard(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    version: str,
    url: str,
    yes: bool = False,
):
    subpath = "{}/{}/{}s/{}".format(owner, project_name, kind, version)

    artifact_url = get_dashboard_url(subpath=subpath)
    if url:
        Printer.print_header("The dashboard is available at: {}".format(artifact_url))
        sys.exit(0)
    if yes or click.confirm(
        "Dashboard page will now open in your browser. Continue?",
        default=True,
    ):
        click.launch(artifact_url)


def pull_project_version(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    version: str,
    path: str,
    download_artifacts: bool = True,
):
    fqn_version = get_versioned_entity_full_name(owner, project_name, version)
    polyaxon_client = ProjectClient(owner=owner, project=project_name)

    try:
        Printer.print_header(
            "Pulling {} version [white]`{}`[/white] ...".format(kind, fqn_version),
        )
        path = polyaxon_client.pull_version(
            kind,
            version,
            path=path,
            download_artifacts=download_artifacts,
        )
        Printer.print_success(
            "Finished pulling the {} version `{}` to `{}`".format(
                kind, fqn_version, path
            )
        )
    except (ApiException, HTTPError) as e:
        handle_cli_error(
            e,
            message="Could not pull the {} version `{}`".format(kind, fqn_version),
        )


def pull_one_or_many_project_versions(
    owner: str,
    project_name: str,
    kind: V1ProjectVersionKind,
    version: str,
    all_versions: bool = None,
    query: str = None,
    limit: int = None,
    offset: int = None,
    path: str = None,
    download_artifacts: bool = True,
):
    def _pull(version_name: str):
        pull_project_version(
            owner=owner,
            project_name=project_name,
            kind=kind,
            version=version_name,
            path=path,
            download_artifacts=download_artifacts,
        )

    if all_versions or any([query, limit, offset]):
        limit = 1000 if all_versions else limit
        versions = list_project_versions_response(
            owner=owner,
            project_name=project_name,
            kind=kind,
            limit=limit,
            offset=offset,
            query=query,
        ).results
        Printer.print_header(f"Pulling {kind} versions (total: {len(versions)}) ...")
        for idx, version in enumerate(versions):
            Printer.print_heading(f"Pulling version {idx + 1}/{len(versions)} ...")
            _pull(version.name)
    elif version:
        _pull(version)
    else:
        Printer.print_error(
            "Please provide a version name, provide a query to filter versions to pull, "
            "or pass the flag `-a/--all` to pull versions.",
            sys_exit=True,
        )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.polyflow.early_stopping.policies import (
    DiffStoppingPolicySchema,
    FailureEarlyStoppingSchema,
    MedianStoppingPolicySchema,
    MetricEarlyStoppingSchema,
    StoppingPolicySchema,
    TruncationStoppingPolicySchema,
    V1DiffStoppingPolicy,
    V1FailureEarlyStopping,
    V1MedianStoppingPolicy,
    V1MetricEarlyStopping,
    V1TruncationStoppingPolicy,
)
from polyaxon.schemas.base import BaseOneOfSchema


class EarlyStoppingSchema(BaseOneOfSchema):
    TYPE_FIELD = "kind"
    TYPE_FIELD_REMOVE = False
    SCHEMAS = {
        V1MetricEarlyStopping.IDENTIFIER: MetricEarlyStoppingSchema,
        V1FailureEarlyStopping.IDENTIFIER: FailureEarlyStoppingSchema,
    }



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Union

from polyaxon.connections.reader import read_keys


def get_aws_access_key_id(
    keys: Optional[Union[str, List[str]]] = None,
    context_path: Optional[str] = None,
    **kwargs,
):
    value = (
        kwargs.get("access_key_id")
        or kwargs.get("aws_access_key_id")
        or kwargs.get("AWS_ACCESS_KEY_ID")
    )
    if value:
        return value
    keys = keys or ["AWS_ACCESS_KEY_ID"]
    return read_keys(context_path=context_path, keys=keys)


def get_aws_secret_access_key(
    keys: Optional[Union[str, List[str]]] = None,
    context_path: Optional[str] = None,
    **kwargs,
):
    value = (
        kwargs.get("secret_access_key")
        or kwargs.get("aws_secret_access_key")
        or kwargs.get("AWS_SECRET_ACCESS_KEY")
    )
    if value:
        return value
    keys = keys or ["AWS_SECRET_ACCESS_KEY"]
    return read_keys(context_path=context_path, keys=keys)


def get_aws_security_token(
    keys: Optional[Union[str, List[str]]] = None,
    context_path: Optional[str] = None,
    **kwargs,
):
    value = (
        kwargs.get("session_token")
        or kwargs.get("aws_session_token")
        or kwargs.get("security_token")
        or kwargs.get("aws_security_token")
        or kwargs.get("AWS_SECURITY_TOKEN")
    )
    if value:
        return value
    keys = keys or ["AWS_SECURITY_TOKEN"]
    return read_keys(context_path=context_path, keys=keys)


def get_region(
    keys: Optional[Union[str, List[str]]] = None,
    context_path: Optional[str] = None,
    **kwargs,
):
    value = (
        kwargs.get("region")
        or kwargs.get("region_name")
        or kwargs.get("aws_region")
        or kwargs.get("AWS_REGION")
    )
    if value:
        return value
    keys = keys or ["AWS_REGION"]
    return read_keys(context_path=context_path, keys=keys)


def get_endpoint_url(
    keys: Optional[Union[str, List[str]]] = None,
    context_path: Optional[str] = None,
    **kwargs,
):
    value = (
        kwargs.get("endpoint_url")
        or kwargs.get("aws_endpoint_url")
        or kwargs.get("AWS_ENDPOINT_URL")
    )
    if value:
        return value
    keys = keys or ["AWS_ENDPOINT_URL"]
    return read_keys(context_path=context_path, keys=keys)


def get_aws_use_ssl(
    keys: Optional[Union[str, List[str]]] = None,
    context_path: Optional[str] = None,
    **kwargs,
):
    value = (
        kwargs.get("use_ssl") or kwargs.get("aws_use_ssl") or kwargs.get("AWS_USE_SSL")
    )
    if value is not None:
        return value
    keys = keys or ["AWS_USE_SSL"]
    value = read_keys(context_path=context_path, keys=keys)
    if value is not None:
        return value
    return True


def get_aws_verify_ssl(
    keys: Optional[Union[str, List[str]]] = None,
    context_path: Optional[str] = None,
    **kwargs,
):
    value = kwargs.get(
        "verify_ssl",
        kwargs.get("aws_verify_ssl", kwargs.get("AWS_VERIFY_SSL", None)),
    )
    if value is not None:
        return value
    keys = keys or ["AWS_VERIFY_SSL"]
    value = read_keys(context_path=context_path, keys=keys)
    if value is not None:
        return value
    return True


def get_aws_legacy_api(
    keys: Optional[Union[str, List[str]]] = None,
    context_path: Optional[str] = None,
    **kwargs,
):
    value = (
        kwargs.get("legacy_api")
        or kwargs.get("aws_legacy_api")
        or kwargs.get("AWS_LEGACY_API")
    )
    if value:
        return value
    keys = keys or ["AWS_LEGACY_API"]
    return read_keys(context_path=context_path, keys=keys)


def get_legacy_api(legacy_api=False, **kwargs):
    legacy_api = legacy_api or get_aws_legacy_api(**kwargs)
    return legacy_api


def get_aws_session(
    context_path=None,
    **kwargs,
):
    import boto3

    aws_access_key_id = get_aws_access_key_id(context_path=context_path, **kwargs)
    aws_secret_access_key = get_aws_secret_access_key(
        context_path=context_path, **kwargs
    )
    aws_session_token = get_aws_security_token(context_path=context_path, **kwargs)
    region_name = get_region(context_path=context_path, **kwargs)
    return boto3.session.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        region_name=region_name,
    )


def get_aws_client(
    client_type,
    context_path=None,
    **kwargs,
):
    session = get_aws_session(
        context_path=context_path,
        **kwargs,
    )
    endpoint_url = get_endpoint_url(context_path=context_path, **kwargs)
    aws_use_ssl = get_aws_use_ssl(context_path=context_path, **kwargs)
    aws_verify_ssl = get_aws_verify_ssl(context_path=context_path, **kwargs)
    return session.client(
        client_type,
        endpoint_url=endpoint_url,
        use_ssl=aws_use_ssl,
        verify=aws_verify_ssl,
    )


def get_aws_resource(
    resource_type,
    context_path=None,
    **kwargs,
):
    session = get_aws_session(
        context_path=context_path,
        **kwargs,
    )
    endpoint_url = get_endpoint_url(context_path=context_path, **kwargs)
    aws_use_ssl = get_aws_use_ssl(context_path=context_path, **kwargs)
    aws_verify_ssl = get_aws_verify_ssl(context_path=context_path, **kwargs)
    return session.resource(
        resource_type,
        endpoint_url=endpoint_url,
        use_ssl=aws_use_ssl,
        verify=aws_verify_ssl,
    )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, List, Union

from polyaxon.utils.list_utils import to_list
from polyaxon.utils.path_utils import get_files_in_path


def hash_value(value: Any, hash_length: int = 12) -> str:
    import hashlib

    return hashlib.md5(str(value).encode("utf-8")).hexdigest()[:hash_length]


def hash_file(
    filepath: str,
    hash_length: int = 12,
    chunk_size: int = 64 * 1024,
    hash_md5: Any = None,
    digest: bool = True,
) -> Union[any, str]:
    import hashlib

    hash_md5 = hash_md5 or hashlib.md5()
    with open(filepath, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()[:hash_length] if digest else hash_md5


def hash_files(
    filepaths: List[str],
    hash_length: int = 12,
    chunk_size: int = 64 * 1024,
    hash_md5: Any = None,
    digest: bool = True,
) -> Union[any, str]:
    import hashlib

    hash_md5 = hash_md5 or hashlib.md5()
    filepaths = to_list(filepaths, check_none=True)
    for filepath in filepaths:
        hash_md5 = hash_file(
            filepath=filepath,
            hash_length=hash_length,
            chunk_size=chunk_size,
            hash_md5=hash_md5,
            digest=False,
        )
    return hash_md5.hexdigest()[:hash_length] if digest else hash_md5


def hash_dir(
    dirpath: str,
    exclude: List[str] = None,
    hash_length: int = 12,
    chunk_size: int = 64 * 1024,
    hash_md5: Any = None,
    digest: bool = True,
) -> Union[any, str]:
    filepaths = get_files_in_path(path=dirpath, exclude=exclude)
    return hash_files(
        filepaths=filepaths,
        hash_length=hash_length,
        chunk_size=chunk_size,
        hash_md5=hash_md5,
        digest=digest,
    )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from marshmallow import ValidationError, fields, validates_schema

from polyaxon.schemas.base import BaseCamelSchema, BaseConfig
from polyaxon.utils.signal_decorators import check_partial


def validate_security_context(user, group):
    if any([user, group]) and not all([user, group]):
        raise ValidationError(
            "Security context requires both `user` and `group` or none."
        )


class SecurityContextSchema(BaseCamelSchema):
    enabled = fields.Bool(allow_none=True)
    run_as_user = fields.Int(allow_none=True)
    run_as_group = fields.Int(allow_none=True)
    fs_group = fields.Int(allow_none=True)
    fs_group_change_policy = fields.Str(allow_none=True)
    allow_privilege_escalation = fields.Bool(allow_none=True)
    run_as_non_root = fields.Bool(allow_none=True)

    @staticmethod
    def schema_config():
        return SecurityContextConfig

    @validates_schema
    @check_partial
    def validate_security_context(self, data, **kwargs):
        validate_security_context(data.get("run_as_user"), data.get("run_as_group"))


class SecurityContextConfig(BaseConfig):
    SCHEMA = SecurityContextSchema
    REDUCED_ATTRIBUTES = [
        "enabled",
        "runAsUser",
        "runAsGroup",
        "fsGroup",
        "fsGroupChangePolicy",
        "allowPrivilegeEscalation",
        "runAsNonRoot",
    ]

    def __init__(
        self,
        enabled=None,
        run_as_user=None,
        run_as_group=None,
        fs_group=None,
        fs_group_change_policy=None,
        allow_privilege_escalation=None,
        run_as_non_root=None,
    ):
        validate_security_context(run_as_user, run_as_group)
        self.enabled = enabled
        self.run_as_user = run_as_user
        self.run_as_group = run_as_group
        self.fs_group = fs_group
        self.fs_group_change_policy = fs_group_change_policy
        self.allow_privilege_escalation = allow_privilege_escalation
        self.run_as_non_root = run_as_non_root



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from polyaxon.env_vars.keys import EV_KEYS_AGENT_INSTANCE, EV_KEYS_ARTIFACTS_STORE_NAME
from polyaxon.exceptions import PolyaxonAgentError


def get_agent_info(agent_instance: str = None):
    agent_instance = agent_instance or os.getenv(EV_KEYS_AGENT_INSTANCE, None)
    if not agent_instance:
        raise PolyaxonAgentError(
            "Could get agent info, "
            "please make sure that this agent was registered in Polyaxon."
        )

    parts = agent_instance.split(".")
    if not len(parts) == 3 or parts[1] != "agents":
        raise PolyaxonAgentError(
            "agent instance is invalid `{}`, "
            "please make sure that this agent was registered in Polyaxon.".format(
                agent_instance
            )
        )
    return parts[0], parts[-1]


def get_artifacts_store_name():
    """Get the artifacts store name"""
    return os.getenv(EV_KEYS_ARTIFACTS_STORE_NAME)



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from polyaxon.k8s.custom_resources.crd import get_custom_object

KIND = "Operation"
PLURAL = "operations"
API_VERSION = "v1"
GROUP = "core.polyaxon.com"


def get_operation_custom_object(
    resource_name: str,
    namespace: str,
    custom_object: Dict,
    annotations: Dict[str, str],
    labels: Dict[str, str],
) -> Dict:
    return get_custom_object(
        resource_name=resource_name,
        namespace=namespace,
        kind=KIND,
        api_version="{}/{}".format(GROUP, API_VERSION),
        labels=labels,
        annotations=annotations,
        custom_object=custom_object,
    )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shlex

from subprocess import PIPE

from psutil import Popen


def run_command(cmd, data, location, chw, env=None):
    cmd_env = None
    if env:
        cmd_env = os.environ.copy()
        cmd_env.update(env)
    cwd = os.getcwd()
    if location is not None and chw is True:
        cwd = location
    elif location is not None and chw is False:
        cmd = "{0} {1}".format(cmd, location)
    r = Popen(
        shlex.split(cmd), stdout=PIPE, stdin=PIPE, stderr=PIPE, cwd=cwd, env=cmd_env
    )
    if data is None:
        output = r.communicate()[0].decode("utf-8")
    else:
        output = r.communicate(input=data)[0]
    return output



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import shutil
import tarfile
import tempfile

from contextlib import contextmanager
from typing import List, Tuple

from polyaxon.exceptions import PolyaxonPathException
from polyaxon.logger import logger
from polyaxon.utils import cli_constants
from polyaxon.utils.list_utils import to_list


def check_or_create_path(path: str = None, is_dir=False) -> None:
    if not is_dir:
        path = os.path.dirname(os.path.abspath(path))
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)


def delete_path(path: str) -> None:
    if not os.path.exists(path):
        return
    try:
        if os.path.isfile(path):
            os.remove(path)
        else:
            shutil.rmtree(path)
    except OSError:
        logger.warning("Could not delete path `%s`", path)


def create_path(path: str) -> None:
    try:
        os.mkdir(path)
    except FileExistsError:
        pass
    except OSError as e:
        logger.warning("Could not create path `%s`, exception %s", path, e)


def get_tmp_path(path: str) -> str:
    return os.path.join("/tmp", path)


def create_tmp_dir(dir_name: str) -> None:
    create_path(get_tmp_path(dir_name))


def delete_tmp_dir(dir_name: str) -> None:
    delete_path(get_tmp_path(dir_name))


def copy_to_tmp_dir(path: str, dir_name: str) -> str:
    tmp_path = get_tmp_path(dir_name)
    if os.path.exists(tmp_path):
        return tmp_path
    try:
        shutil.copytree(path, tmp_path)
    except FileExistsError as e:
        logger.warning("Path already exists `%s`, exception %s", path, e)
    return tmp_path


def copy_file(filename, path_to, use_basename=True):
    if use_basename:
        path_to = append_basename(path_to, filename)

    if filename == path_to:
        return

    check_or_create_path(path_to, is_dir=False)
    shutil.copy(filename, path_to)
    return path_to


@contextmanager
def get_files_by_paths(file_type, filepaths):
    local_files = []
    total_file_size = 0

    for filepath in filepaths:
        local_files.append(
            (file_type, (unix_style_path(filepath), open(filepath, "rb"), "text/plain"))
        )
        total_file_size += os.path.getsize(filepath)

    yield local_files, total_file_size

    # close all files to avoid WindowsError: [Error 32]
    for f in local_files:
        f[1][1].close()


def get_files_and_dirs_in_path(
    path: str,
    exclude: List[str] = None,
    collect_dirs: bool = False,
) -> Tuple[List[str], List[str]]:
    result_files = []
    result_dirs = []
    exclude = to_list(exclude, check_none=True)
    for root, dirs, files in os.walk(path, topdown=True):
        if exclude:
            dirs[:] = [d for d in dirs if d not in exclude]
        logger.debug("Root:%s, Dirs:%s", root, dirs)
        for file_name in files:
            result_files.append(os.path.join(root, file_name))
        if collect_dirs:
            for dir_name in dirs:
                result_dirs.append(os.path.join(root, dir_name))
    return result_files, result_dirs


def get_files_in_path(path: str, exclude: List[str] = None) -> List[str]:
    return get_files_and_dirs_in_path(path, exclude, False)[0]


def get_dirs_under_path(path: str) -> List[str]:
    return [
        name for name in os.listdir(path) if os.path.isdir(os.path.join(path, name))
    ]


@contextmanager
def get_files_in_path_context(path: str, exclude: List[str] = None):
    """
    Gets all the files under a certain path.

    Args:
        path: `str`. The path to traverse for collecting files.
        exclude: `list`. List of paths to excludes.

    Returns:
         list of files collected under the path.
    """
    yield get_files_in_path(path, exclude=exclude)


def unix_style_path(path):
    if os.path.sep != "/":
        return path.replace(os.path.sep, "/")
    return path


def create_init_file():
    if os.path.exists(cli_constants.INIT_FILE_PATH):
        return False

    with open(cli_constants.INIT_FILE_PATH, "w") as f:
        f.write(cli_constants.INIT_FILE_TEMPLATE)

    return True


def create_tarfile(files: List[str], tar_path: str, relative_to: str = None) -> None:
    """Create a tar file based on the list of files passed"""
    with tarfile.open(tar_path, "w:gz") as tar:
        for f in files:
            arcname = os.path.relpath(f, relative_to) if relative_to else None
            tar.add(f, arcname=arcname)


@contextmanager
def create_tarfile_from_path(files, path_name, relative_to: str = None):
    """Create a tar file based on the list of files passed"""
    fd, filename = tempfile.mkstemp(prefix=path_name, suffix=".tar.gz")
    create_tarfile(files, filename, relative_to)
    yield filename

    # clear
    os.close(fd)
    os.remove(filename)


def untar_file(
    filename: str = None,
    delete_tar: bool = True,
    extract_path: str = None,
    use_filepath: bool = True,
):
    extract_path = extract_path or "."
    if use_filepath:
        extract_path = os.path.join(extract_path, filename.split(".tar.gz")[0])
    check_or_create_path(extract_path, is_dir=True)
    logger.info("Untarring the contents of the file ...")
    # Untar the file
    with tarfile.open(filename) as tar:
        tar.extractall(extract_path)
    if delete_tar:
        logger.info("Cleaning up the tar file ...")
        os.remove(filename)
    return extract_path


def move_recursively(src, dst):
    files = os.listdir(src)

    for f in files:
        shutil.move(os.path.join(src, f), dst)


def append_basename(path, filename):
    """
    Adds the basename of the filename to the path.

    Args:
        path: `str`. The path to append the basename to.
        filename: `str`. The filename to extract the base name from.

    Returns:
         str
    """
    return os.path.join(path, os.path.basename(filename))


def check_dirname_exists(path, is_dir=False):
    if not is_dir:
        path = os.path.dirname(os.path.abspath(path))
    if not os.path.isdir(path):
        raise PolyaxonPathException(
            "The parent path is not a directory {}".format(path)
        )


def create_polyaxon_tmp():
    base_path = os.path.join("/tmp", ".polyaxon")
    if not os.path.exists(base_path):
        try:
            os.makedirs(base_path)
        except OSError:
            # Except permission denied and potential race conditions
            # in multi-threaded environments.
            logger.warning("Could not create config directory `%s`", base_path)
    return base_path


def get_path_extension(filepath: str):
    return ".".join(os.path.basename(filepath).split(".")[1:]).lower()


def get_base_filename(filepath: str):
    return os.path.basename(filepath).split(".")[0]


def module_type(obj, type_pattern):
    obj_type = type(obj)
    module = obj_type.__module__
    name = obj_type.__name__
    actual_fqn = "%s.%s" % (module, name)
    if isinstance(type_pattern, str):
        return type_pattern == actual_fqn
    else:
        return type_pattern.match(actual_fqn) is not None


def copy_file_path(from_path: str, asset_path: str):
    check_or_create_path(asset_path, is_dir=False)
    shutil.copy(from_path, asset_path)


def copy_dir_path(from_path: str, asset_path: str):
    check_or_create_path(asset_path, is_dir=False)
    shutil.copytree(from_path, asset_path)


def copy_file_or_dir_path(from_path: str, asset_path: str, use_basename: bool = False):
    if use_basename:
        dir_name = os.path.basename(os.path.normpath(from_path))
        asset_path = (
            os.path.join(asset_path, dir_name) if asset_path is not None else dir_name
        )
    check_or_create_path(asset_path, is_dir=False)
    if os.path.isfile(from_path):
        try:
            shutil.copy(from_path, asset_path)
        except shutil.SameFileError:
            pass
    else:
        shutil.copytree(from_path, asset_path)

    return asset_path


def get_relative_path_to(base_path, paths: List[str]) -> List[str]:
    results = []
    if not paths:
        return results
    for d in paths:
        if not os.path.isabs(d):
            results.append(os.path.join(base_path, d))
        else:
            results.append(d)

    return results



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from typing import List

from polyaxon import settings
from polyaxon.utils.path_utils import create_tarfile, get_files_in_path


def tar_dir(download_path: str) -> str:
    outputs_files = get_files_in_path(download_path)
    tar_base_name = os.path.basename(download_path)
    tar_name = "{}.tar.gz".format(tar_base_name)
    target_path = os.path.join(settings.CLIENT_CONFIG.archive_root, tar_name)
    create_tarfile(files=outputs_files, tar_path=target_path, relative_to=download_path)
    return target_path


def tar_files(
    download_name: str, outputs_files: List[str], relative_to: str = None
) -> str:
    tar_name = "{}.tar.gz".format(download_name)
    target_path = os.path.join(settings.CLIENT_CONFIG.archive_root, tar_name)
    create_tarfile(files=outputs_files, tar_path=target_path, relative_to=relative_to)
    return target_path



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from marshmallow import fields

from polyaxon.schemas.base import BaseCamelSchema, BaseConfig


class SSLSchema(BaseCamelSchema):
    enabled = fields.Bool(allow_none=True)
    secret_name = fields.Str(allow_none=True)
    path = fields.Str(allow_none=True)

    @staticmethod
    def schema_config():
        return SSLConfig


class SSLConfig(BaseConfig):
    SCHEMA = SSLSchema
    REDUCED_ATTRIBUTES = ["enabled"]

    def __init__(self, enabled=None, secret_name=None, path=None):  # noqa
        self.enabled = enabled
        self.secret_name = secret_name
        self.path = path



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from polyaxon.polypod.custom_resources.job import get_job_custom_resource
from polyaxon.polypod.custom_resources.kubeflow import (
    get_mpi_job_custom_resource,
    get_mx_job_custom_resource,
    get_pytorch_job_custom_resource,
    get_tf_job_custom_resource,
    get_xgb_job_custom_resource,
)
from polyaxon.polypod.custom_resources.service import get_service_custom_resource



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

NEWLINES = ("\n", "\r", "\r\n")

INIT_COMMAND = (
    "`polyaxon init -p PROJECT_NAME [--polyaxonfile] [--git-connection] [--git-url]`"
)

DEFAULT_IGNORE_LIST = """
./.polyaxon
.git
.eggs
eggs
lib
lib64
parts
sdist
var
*.pyc
*.swp
.DS_Store
__pycache__/
.pytest_cache/
*.py[cod]
*$py.class
.mypy_cache/
.vscode/
.mr.developer.cfg
.pydevproject
.project
.settings/
.idea/
.DS_Store
# temp files
*~
# C extensions
*.so
# Distribution / packaging
.Python
dist/
pydist/
*.egg-info/
.installed.cfg
*.egg
# PyInstaller
*.manifest
*.spec
# IPython Notebook
.ipynb_checkpoints
# pyenv
.python-version
"""

INIT_FILE_PATH = "polyaxonfile.yaml"

INIT_FILE_TEMPLATE = """---
version: 1.1
kind: component
run:
  kind: job
  container:
    # image: # image to use
    # command: # Command to use
"""

INIT_FILE = "init"
DEBUG_FILE = "debug"



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

import click

from polyaxon.exceptions import PolyaxonException


@click.command()
@click.argument("component")
@click.option(
    "--path", default="./web", help="Path where the config should be generated."
)
@click.option("--root", help="Absolute root where the configs, default to pwd")
def proxy(component, path, root):
    """Create api proxy."""
    from polyaxon.proxies.generators import (
        generate_api_conf,
        generate_forward_proxy_cmd,
        generate_gateway_conf,
        generate_streams_conf,
    )
    from polyaxon.services.values import PolyaxonServices
    from polyaxon.settings import set_proxies_config

    if not root:
        root = os.path.abspath(".")

    set_proxies_config()

    if PolyaxonServices.is_api(component):
        generate_api_conf(path=path, root=root)
    elif PolyaxonServices.is_streams(component):
        generate_streams_conf(path=path, root=root)
    elif PolyaxonServices.is_gateway(component):
        generate_gateway_conf(path=path, root=root)
        generate_forward_proxy_cmd(path=path)
    else:
        raise PolyaxonException("Component {} is not recognized".format(component))



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List

from marshmallow import fields, validate

import polyaxon_sdk

from polyaxon.exceptions import PolyaxonSchemaError
from polyaxon.polyflow.io import V1IO, IOSchema
from polyaxon.polyflow.operations.base import BaseOp, BaseOpSchema
from polyaxon.polyflow.params import ParamSpec, ops_params
from polyaxon.polyflow.run import RunMixin, RunSchema


class CompiledOperationSchema(BaseOpSchema):
    kind = fields.Str(allow_none=True, validate=validate.Equal("compiled_operation"))
    inputs = fields.List(fields.Nested(IOSchema), allow_none=True)
    outputs = fields.List(fields.Nested(IOSchema), allow_none=True)
    contexts = fields.List(fields.Nested(IOSchema), allow_none=True)
    run = fields.Nested(RunSchema, required=True)

    @staticmethod
    def schema_config():
        return V1CompiledOperation


class V1CompiledOperation(BaseOp, RunMixin, polyaxon_sdk.V1CompiledOperation):
    SCHEMA = CompiledOperationSchema
    IDENTIFIER = "compiled_operation"
    REDUCED_ATTRIBUTES = BaseOp.REDUCED_ATTRIBUTES + [
        "inputs",
        "outputs",
        "contexts",
        "run",
    ]

    def get_run_kind(self):
        return self.run.kind if self.run else None

    def validate_params(
        self,
        params: Dict = None,
        context: Dict = None,
        is_template: bool = True,
        check_all_refs: bool = False,
        parse_values: bool = False,
        parse_joins: bool = True,
    ) -> List[ParamSpec]:
        return ops_params.validate_params(
            inputs=self.inputs,
            outputs=self.outputs,
            contexts=self.contexts,
            params=params,
            matrix=self.matrix,
            joins=self.joins if parse_joins else None,
            context=context,
            is_template=is_template,
            check_all_refs=check_all_refs,
            parse_values=parse_values,
        )

    def apply_params(self, params=None, context=None):
        context = context or {}
        validated_params = self.validate_params(
            params=params,
            context=context,
            is_template=False,
            check_all_refs=True,
            parse_values=True,
        )
        if not validated_params:
            return

        param_specs = {}
        for param in validated_params:
            param_specs[param.name] = param

        processed_params = set([])

        def set_io(io):
            if not io:
                return
            for i in io:
                if i.name in param_specs:
                    processed_params.add(i.name)
                    i.is_optional = True
                    if param_specs[i.name].param.is_literal:
                        current_param = param_specs[i.name].param
                        value = current_param.value
                        if hasattr(value, "to_param"):
                            value = value.to_param()
                        i.value = value
                        if current_param.connection:
                            i.connection = current_param.connection
                        if current_param.to_init:
                            i.to_init = current_param.to_init
                        if current_param.to_env:
                            i.to_env = current_param.to_env

        def set_contexts() -> List[V1IO]:
            context_params = [p for p in param_specs if p not in processed_params]
            contexts = []
            for p in context_params:
                current_param = param_specs[p].param
                contexts.append(
                    V1IO(
                        name=p,
                        value=current_param.value,
                        is_optional=True,
                        connection=current_param.connection,
                        to_init=current_param.to_init,
                        to_env=current_param.to_env,
                    )
                )

            return contexts

        set_io(self.inputs)
        set_io(self.outputs)
        self.contexts = set_contexts()

    def get_env_io(self) -> List[List]:
        def get_env_io(io: List[V1IO]) -> List[List]:
            return [[i.to_env, i.value] for i in io if i.to_env]

        return (
            get_env_io(self.inputs or [])
            + get_env_io(self.outputs or [])
            + get_env_io(self.contexts or [])
        )

    def get_io_names(self) -> List[str]:
        def get_io_name(io: List[V1IO]) -> List[str]:
            return [i.name for i in io if i.name]

        return (
            get_io_name(self.inputs or [])
            + get_io_name(self.outputs or [])
            + get_io_name(self.contexts or [])
        )

    def apply_image_destination(self, image: str):
        self.run.apply_image_destination(image)

    @property
    def has_pipeline(self):
        return self.is_dag_run or self.matrix or self.schedule

    def validate_build(self):
        if self.build and self.is_dag_run:
            raise PolyaxonSchemaError(
                "Operations with dag runtime do not support the `build` section."
            )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from typing import Dict

from polyaxon.client.init import get_client_or_raise
from traceml.artifacts import V1ArtifactKind


def create_file_lineage(filepath: str, summary: Dict, kind: str):
    if not filepath:
        return
    filename = os.path.basename(filepath)
    if not kind:
        if "dockerfile" in filename.lower():
            kind = V1ArtifactKind.DOCKERFILE
        else:
            kind = V1ArtifactKind.FILE

    run_client = get_client_or_raise()
    if not run_client:
        return

    run_client.log_artifact_ref(
        path=filepath,
        kind=kind,
        name=filename,
        summary=summary,
        is_input=True,
    )



#!/usr/bin/python
#
# Copyright 2018-2022 Polyaxon, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand


def read_readme():
    if not os.path.exists("../README.md"):
        return ""
    with open("../README.md") as f:
        return f.read()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


with open(os.path.join("./polyaxon/pkg.py"), encoding="utf8") as f:
    pkg = {}
    exec(f.read(), pkg)

with open("requirements/requirements.txt") as requirements_file:
    requirements = requirements_file.read().splitlines()

if os.environ.get("USE_PROD_PACKAGES"):
    with open("requirements/prod.txt") as requirements_file:
        requirements += requirements_file.read().splitlines()

with open("requirements/dev.txt") as requirements_file:
    dev_requirements = requirements_file.read().splitlines()

extra = {
    "dev": dev_requirements,
    "async": ["aiohttp >= 3.0.0"],
    "gcs": ["gcsfs"],
    "s3": ["s3fs"],
    "azure": ["adlfs"],
    "docker": ["docker"],
    "git": ["gitpython"],
    "numpy": ["numpy"],
    "sandbox": ["polyaxon-deploy"],
}

setup(
    name=pkg["NAME"],
    version=pkg["VERSION"],
    description=pkg["DESC"],
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    maintainer=pkg["AUTHOR"],
    maintainer_email=pkg["EMAIL"],
    author=pkg["AUTHOR"],
    author_email=pkg["EMAIL"],
    url=pkg["URL"],
    license=pkg["LICENSE"],
    platforms="any",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    keywords=[
        "polyaxon",
        "aws",
        "s3",
        "microsoft",
        "azure",
        "google cloud storage",
        "gcs",
        "deep-learning",
        "machine-learning",
        "data-science",
        "neural-networks",
        "artificial-intelligence",
        "ai",
        "reinforcement-learning",
        "kubernetes",
        "aws",
        "microsoft",
        "azure",
        "google cloud",
        "tensorFlow",
        "pytorch",
    ],
    install_requires=requirements,
    extras_require=extra,
    entry_points={
        "console_scripts": ["polyaxon = polyaxon.main:cli", "plx = polyaxon.main:cli"]
    },
    python_requires=">=3.5",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    tests_require=["pytest"],
    cmdclass={"test": PyTest},
)
